import os
import sys

import click
import requests
from loguru import logger
from nornir_rich.functions import print_result
from nornir_rich.progress_bar import RichProgressBar
from nornir_utils.plugins.functions import print_title

from netbox_network_importer.config import setup_logger
from netbox_network_importer.connections.Netbox import Netbox
from netbox_network_importer.connections.Nornir import Nornir
from netbox_network_importer.connections.PyAts import PyAtsNetbox
from netbox_network_importer.helper import remove_key_from_results
from netbox_network_importer.outputers.output import html_output, json_output
from netbox_network_importer.processors.save_netbox_results import SaveNetboxResults
from netbox_network_importer.tasks.connection_test import connection_test
from netbox_network_importer.tasks.synchronizer import (
    synchronize_interfaces,
    synchronize_interfaces_bandwidths,
    synchronize_interfaces_ips,
    synchronize_interfaces_lags,
    synchronize_interfaces_vlans,
    synchronize_serials,
)


@click.group()
@click.option(
    "--configs",
    "-c",
    type=click.Path(),
    multiple=False,
    help="path to folder with configurations",
)
def cli(configs):
    # set custom configs folder path
    if configs:
        os.environ["CONFIG_FOLDER_PATH"] = os.path.abspath(configs)

    setup_logger()


@cli.command()
@click.option(
    "--device",
    "-d",
    type=str,
    multiple=False,
    required=True,
    help="Run on specificied device",
)
@click.option(
    "--command",
    "-c",
    type=str,
    multiple=False,
    required=True,
    help="Run command e.g. show version",
)
def pyats(device, command):
    """
    Connect to device using pyats and pynetbox and print parsed command
    """

    pyats = PyAtsNetbox()
    nb = Netbox.connect()

    nb_dev = nb.dcim.devices.get(name=device)
    hostname = nb_dev.name

    dev = pyats.connect_device(hostname)

    from pprint import pprint

    pprint(dev.parse(command))


@cli.command()
@click.option(
    "--devices", "-d", type=str, multiple=True, help="Run on specificied devices"
)
@click.option(
    "--platforms", "-p", type=str, multiple=True, help="Run on specificied platforms"
)
@click.option("--no-progress-bar", is_flag=True)
@logger.catch
# CLI callable command
def synchronize(devices, platforms, no_progress_bar):
    """Run set of Nornir task to update data in Netbox from network devices

    Args:
        devices (str): device name filter, can be used multiple times
        platforms (str): platform name filter, can be used multiple times
        no_progress_bar (bool): use to hide progress bar
    Returns:
        dict: Dictionary of hosts, it's actions and results
    """

    result_data = run_synchronizer(devices, platforms, no_progress_bar)
    return result_data


@cli.command()
@click.option('--devices', '-d',  type=str, multiple=True, help="Run on specificied devices", required=True)
@logger.catch
# CLI callable command
def synchronize_device(devices):
    """Synchronize specific device data in Netbox from network

    Args:
        devices (str): device name filter, can be used multiple times
    Returns:
        dict: Dictionary of hosts, it's actions and results
    """

    result_data = run_synchronizer(devices, [], False)
    return result_data


def run_synchronizer(devices, platforms, no_progress_bar=True):
    """Run set of Nornir task to update data in Netbox from network devices

    Args:
        devices (str): device name filter, can be used multiple times
        platforms (str): platform name filter, can be used multiple times
        no_progress_bar (bool): use to hide progress bar
    Returns:
        dict: Dictionary of hosts, it's actions and results
    """
    # store netbox process results into a dict
    save_netbox_results_data = {}

    # init devices (nornir) with Netbox
    # filter out devices without IP or Platform
    try:
        nr = Nornir.init_nornir_w_netbox(
            filter_parameters={
                "name": devices,
                "platform": platforms,
                "has_primary_ip": True,
                "platform_id__n": "null",
                "status": "active",
            }
        )
    except ValueError as e:
        logger.error(f"Error connecting to Netbox: {e}")
        logger.error("Exiting...")
        sys.exit(1)
    except requests.exceptions.ConnectionError as e:
        logger.error(f"Error connecting to Netbox: {e}")
        logger.error("Exiting...")
        sys.exit(1)

    # init device connector with pyAts via Netbox
    # filter out devices without IP or Platform
    try:
        nb_pyats_devices = PyAtsNetbox(
            url_filter="platform_id__n=null&has_primary_ip=True"
        )
    except KeyError as e:
        logger.error(
            f"Unable to create PyATS testbed structure through Netbox: {e}")
        logger.error("Exiting...")
        sys.exit(1)

    print_result_keys = ["name", "result", "diff", "exception"]

    # Connection Test - Skip failed hosts from this action
    print_title("Connection Test")
    nr = nr.with_processors(get_processors(
        save_netbox_results_data, no_progress_bar))

    connection_tests = nr.run(task=connection_test, pyats=nb_pyats_devices)
    print_result(connection_tests, vars=print_result_keys, expand=True)

    skipped_hosts = []
    if connection_tests.failed_hosts:
        # Skip failed hosts in other processing
        for host, result in connection_tests.failed_hosts.items():
            skipped_hosts.append(host)
            # remove host from other processing
            nr.inventory.hosts.pop(host)
        print_title(f"SKIPPING HOSTS: {skipped_hosts}")

    # Create/Update Netbox Interaces
    save_netbox_results_data = run_task(
        task_method=synchronize_serials,
        skipped_hosts=skipped_hosts,
        result_dict_data=save_netbox_results_data,
        no_progress_bar=no_progress_bar,
        print_result_keys=print_result_keys,
        nornir=nr,
        pyats=nb_pyats_devices,
    )

    # Create/Update Netbox Interaces
    save_netbox_results_data = run_task(
        task_method=synchronize_interfaces,
        skipped_hosts=skipped_hosts,
        result_dict_data=save_netbox_results_data,
        no_progress_bar=no_progress_bar,
        print_result_keys=print_result_keys,
        nornir=nr,
        pyats=nb_pyats_devices,
    )

    # CRUD Netbox LAGs
    save_netbox_results_data = run_task(
        task_method=synchronize_interfaces_lags,
        skipped_hosts=skipped_hosts,
        result_dict_data=save_netbox_results_data,
        no_progress_bar=no_progress_bar,
        print_result_keys=print_result_keys,
        nornir=nr,
        pyats=nb_pyats_devices,
    )

    # CRUD netbox IPs on interfaces
    save_netbox_results_data = run_task(
        task_method=synchronize_interfaces_ips,
        skipped_hosts=skipped_hosts,
        result_dict_data=save_netbox_results_data,
        no_progress_bar=no_progress_bar,
        print_result_keys=print_result_keys,
        nornir=nr,
        pyats=nb_pyats_devices,
    )

    # Create VLANs
    save_netbox_results_data = run_task(
        task_method=synchronize_interfaces_vlans,
        skipped_hosts=skipped_hosts,
        result_dict_data=save_netbox_results_data,
        no_progress_bar=no_progress_bar,
        print_result_keys=print_result_keys,
        nornir=nr,
        pyats=nb_pyats_devices,
    )

    # Update Netbox Interaces Bandwidths
    save_netbox_results_data = run_task(
        task_method=synchronize_interfaces_bandwidths,
        skipped_hosts=skipped_hosts,
        result_dict_data=save_netbox_results_data,
        no_progress_bar=no_progress_bar,
        print_result_keys=print_result_keys,
        nornir=nr,
        pyats=nb_pyats_devices,
    )

    netbox_result_data_without_not_changed = remove_key_from_results(
        save_netbox_results_data, "NOT_CHANGED"
    )
    json_output(save_netbox_results_data)
    html_output(netbox_result_data_without_not_changed)

    return save_netbox_results_data


def get_processors(netbox_result_data: dict, no_progress_bar=False):
    """Returns list of processors
    Created because of reseting RichProgressBar - if it's
    not reset, then it will print bad data

    Args:
        no_progress_bar (bool, optional): Wheter to use use RichProgressBar or not. Defaults to False.
        netbox_result_data (dict, optional): Dict, which will be filled with data.

    Returns:
        _type_: list of processors
    """
    processors = [SaveNetboxResults(netbox_result_data)]
    if not no_progress_bar:
        processors.append(RichProgressBar())
    return processors


def post_process_skipped_hosts(complete_task, skipped_hosts, data) -> dict:
    for skipped_host in skipped_hosts:
        data[skipped_host][complete_task.name] = {"skipped": True}
    return data


def run_task(
    task_method,
    skipped_hosts,
    result_dict_data,
    no_progress_bar,
    print_result_keys,
    nornir,
    pyats,
):
    # Create/Update Netbox Interaces
    print_title(f"Running {task_method.__name__} - skipping {skipped_hosts}")
    # always call with_processors - otherwise progress bar gets crazy
    nr = nornir.with_processors(get_processors(
        result_dict_data, no_progress_bar))

    nr_run_result = nr.run(task=task_method, on_failed=True, pyats=pyats)
    # add info to complete result, that task was skipped
    result_dict_data = post_process_skipped_hosts(
        complete_task=nr_run_result, skipped_hosts=skipped_hosts, data=result_dict_data
    )
    print_result(nr_run_result, vars=print_result_keys, expand=True)

    return result_dict_data


if __name__ == "__main__":
    # Display CLI
    cli()



from netbox_network_importer.connections.Netbox import Netbox
from netbox_network_importer.helper import canonical_interface_name_edited, get_diff
from netbox_network_importer.results.results import Action, NetboxResult, Status


def process_interfaces(host, parsed_interfaces) -> list:
    """CRUD on netbox interfaces
    - Delete interfaces, that exists in netbox but are not passed from device
    - Update interfaces, which already exists in netbox
    - Create interfaces, which does not exists in netbox

    :param parsed_interfaces: parsed interfaces for netbox operations
    :param host: host of the task (netbox device representation)

    :returns: list of NetboxResult
    """

    RESULTS = []  # list of NetboxResult

    # Init connection to Netbox via PyNetbox
    nb = Netbox.connect()

    # find the device instance in netbox
    dev = nb.dcim.devices.get(host.data["id"])

    # find interfaces linked to the device
    ifcs_filter = nb.dcim.interfaces.filter(device_id=dev.id)

    # convert filtered interfaces into dictionary of pynetbox interface instances
    nb_interfaces_dict = {
        canonical_interface_name_edited(ifc.name): ifc
        for ifc in [ifc for ifc in ifcs_filter]
    }

    # setup NB parameters
    for ifc, properties in parsed_interfaces.items():
        mac_address = properties.get("mac_address")
        ifc_params = {
            "name": ifc,
            "enabled": properties["enabled"],
            "description": properties["description"].strip(),
            "type": properties["type"],
            "mtu": properties["mtu"],
        }

        # get NB interface instance
        nb_ifc = nb_interfaces_dict.get(ifc, None)

        if nb_ifc:
            # Skip IGNORED interfaces
            if nb_ifc.custom_fields.get("ignore_importer", False) == True:
                RESULTS.append(
                    NetboxResult(
                        result=f"{nb_ifc.name} - Ignored by Importer - Skipping",
                        status=Status.SKIPPED,
                        action=Action.LOOKUP,
                        diff="",
                    )
                )

                nb_interfaces_dict.pop(nb_ifc.name, None)
                continue
            else:
                # If interface already exists in netbox, update it
                RESULTS.append(interface_update(nb_ifc=nb_ifc, params=ifc_params))
        else:
            # Otherwise, create it
            ifc_params["device"] = {"id": dev.id}
            RESULTS.append(interface_create(ifc_params=ifc_params, nb=nb))

        # Process MAC address
        nb_ifc = nb.dcim.interfaces.get(name=ifc, device_id=dev.id)
        if nb_ifc:
            if nb_ifc.primary_mac_address:
                RESULTS.append(mac_address_update(nb_ifc, mac_address))
            else:
                RESULTS.append(mac_address_create(nb_ifc, mac_address, nb))

        # Pop device from dictionary if it's updated / created
        nb_interfaces_dict.pop(ifc, None)

    # If any interface is left in `nb_interfaces_dict`, then it should be removed from netbox
    #   - Interface exists in netbox, but the interface was not passed from network
    for ifc_name, nb_ifc in nb_interfaces_dict.items():
        RESULTS.append(interface_delete(ifc=nb_ifc))

    return RESULTS


def interface_delete(ifc):
    try:
        if ifc.cable:
            return NetboxResult(
                result=f"{ifc.name} - CHECK MANUALLY - Cable exists",
                status=Status.CHECK_MANUALLY,
                action=Action.DELETE,
                diff={
                    "name": ifc.name,
                    "device_name": ifc.device.name,
                    "description": ifc.description,
                    "cable_id": ifc.cable.id,
                },
            )

        if ifc.custom_fields.get("ignore_importer", False) == True:
            return NetboxResult(
                result=f"{ifc.name} - CHECK MANUALLY - Inteface Ignored",
                status=Status.CHECK_MANUALLY,
                action=Action.DELETE,
                diff={
                    "name": ifc.name,
                    "device_name": ifc.device.name,
                    "description": ifc.description,
                    "ignore_importer": ifc.custom_fields.get("ignore_importer", False),
                },
            )

        if ifc.delete():
            return NetboxResult(
                result=f"{ifc.name} - deleted successfully",
                status=Status.CHANGED,
                action=Action.DELETE,
                diff={
                    "name": ifc.name,
                    "device": ifc.device.name,
                    "description": ifc.description,
                },
            )
        else:
            return NetboxResult(
                result=f"{ifc.name} - could not be deleted",
                status=Status.ANOMALLY,
                action=Action.DELETE,
                diff={},
            )
    except Exception as e:
        return NetboxResult(
            result=f"{ifc.name} - Exception Occurs: {e}",
            status=Status.EXCEPTION,
            action=Action.DELETE,
            diff={},
            exception=e,
        )


def mac_address_update(nb_ifc, mac_address):
    # Handle MAC address separately since it's now a separate model
    if mac_address:
        # Create MAC address object if it doesn't exist
        try:
            mac_addr_obj = nb_ifc.primary_mac_address
            before = str(mac_addr_obj.mac_address)

            if mac_addr_obj.mac_address != mac_address:
                mac_addr_obj.mac_address = mac_address
                nb_ifc.primary_mac_address = mac_addr_obj
                mac_addr_obj.save()
                nb_ifc.save()

                return NetboxResult(
                    result=f"{nb_ifc.name} - MAC address updated successfully",
                    status=Status.CHANGED,
                    action=Action.UPDATE,
                    diff=get_diff(before, mac_address),
                )
            else:
                return NetboxResult(
                    result=f"{nb_ifc.name} - MAC address already exists",
                    status=Status.NOT_CHANGED,
                    action=Action.UPDATE,
                    diff={},
                )

        except Exception as e:
            # MAC address might already exist or be invalid
            return NetboxResult(
                result=f"{nb_ifc.name} - MAC address could not be updated",
                status=Status.ANOMALLY,
                action=Action.UPDATE,
                diff={},
            )
    else:
        return NetboxResult(
            result=f"{nb_ifc.name} - MAC address not provided",
            status=Status.ANOMALLY,
            action=Action.UNKNOWN,
            diff={},
        )


def mac_address_create(nb_ifc, mac_address, nb):
    if mac_address:
        try:
            mac_addr_obj = nb.dcim.mac_addresses.create(
                assigned_object_id=nb_ifc.id,
                assigned_object_type="dcim.interface",
                mac_address=mac_address,
            )
            nb_ifc.primary_mac_address = mac_addr_obj
            nb_ifc.save()
            return NetboxResult(
                result=f"{nb_ifc.name} - MAC address created successfully",
                status=Status.CHANGED,
                action=Action.CREATE,
                diff={"mac_address": mac_address},
            )
        except Exception as e:
            return NetboxResult(
                result=f"{nb_ifc.name} - MAC address could not be created",
                status=Status.ANOMALLY,
                action=Action.CREATE,
                diff={},
            )
    else:
        return NetboxResult(
            result=f"{nb_ifc.name} - MAC address not provided",
            status=Status.NOT_CHANGED,
            action=Action.CREATE,
            diff={},
        )


def interface_update(nb_ifc, params):
    try:
        # Get data before changes
        before = get_changed_params(nb_ifc)

        # TODO: Don't change ifc type if Other -
        if params.get("type", None) == "other":
            params.pop("type")

        if nb_ifc.update(params):
            # TODO: Reload ifc
            nb_ifc = nb_ifc.api.dcim.interfaces.get(nb_ifc.id)
            after = get_changed_params(nb_ifc)

            return NetboxResult(
                result=f"{nb_ifc.name} - saved successfully",
                status=Status.CHANGED,
                action=Action.UPDATE,
                diff=get_diff(before, after),
            )
        else:
            return NetboxResult(
                result=f"{nb_ifc.name} - nothing to do",
                action=Action.UPDATE,
                status=Status.NOT_CHANGED,
                diff={},
            )
    except Exception as e:
        return NetboxResult(
            result=f"{nb_ifc.name} - Exception Occurs: {e}",
            status=Status.EXCEPTION,
            action=Action.DELETE,
            diff={},
            exception=e,
        )


def interface_create(ifc_params, nb):
    try:
        netbox_interface = nb.dcim.interfaces.create(ifc_params)

        if netbox_interface:
            return NetboxResult(
                result=f"{netbox_interface.name} - created successfully",
                status=Status.CHANGED,
                action=Action.CREATE,
                diff=ifc_params,
            )
        else:
            return NetboxResult(
                result=f"{ifc_params['name']} - ERROR!",
                status=Status.FAILED,
                action=Action.CREATE,
                diff={},
            )
    except Exception as e:
        return NetboxResult(
            result=f"{ifc_params['name']} - Exception Occurs: {e}",
            status=Status.EXCEPTION,
            action=Action.DELETE,
            diff={},
            exception=e,
        )


def get_changed_params(nb_interface):
    return {
        "name": nb_interface.name,
        "enabled": nb_interface.enabled,
        "description": nb_interface.description,
        "type": nb_interface.type,
        "mtu": nb_interface.mtu,
        "mac_address": nb_interface.mac_address,
    }



import logging
import os
import sys

import appdirs
import yaml
from loguru import logger

from netbox_network_importer import __appname__

logging.getLogger("urllib3.connectionpool").setLevel(logging.INFO)
logging.getLogger("pyats").setLevel(logging.WARNING)
logging.getLogger("git").setLevel(logging.INFO)
logging.getLogger("genie").setLevel(logging.INFO)
logging.getLogger("nornir.core").setLevel(logging.WARNING)
logging.getLogger("genie.utils.summary").setLevel(logging.INFO)
logging.getLogger("genie.ops.base.maker").setLevel(logging.INFO)
logging.getLogger("blib2to3.pgen2.driver").setLevel(logging.INFO)
logging.getLogger("paramiko.transport").setLevel(logging.WARNING)
logging.getLogger("netmiko").setLevel(logging.WARNING)
logging.getLogger("napalm").setLevel(logging.WARNING)
logging.getLogger("pyats.contrib.creators.netbox").setLevel(logging.ERROR)

os.environ["XDG_CONFIG_DIRS"] = "/etc"
CONFIG_DIRS = (
    appdirs.user_config_dir(__appname__),
    appdirs.site_config_dir(__appname__),
)
CONFIG_FILENAME = "config.yml"


def get_config():
    """
    Get config file and load it with yaml
    :returns: loaded config in yaml, as a dict object
    """
    if getattr(get_config, "cache", None):
        return get_config.cache

    if os.environ.get("CONFIG_FOLDER_PATH"):
        config_path = os.path.join(os.environ["CONFIG_FOLDER_PATH"], CONFIG_FILENAME)
    else:
        for d in CONFIG_DIRS:
            config_path = os.path.join(d, CONFIG_FILENAME)
            if os.path.isfile(config_path):
                break
    try:
        with open(config_path, "r") as config_file:
            conf = yaml.safe_load(config_file)
            get_config.cache = conf
            return conf
    except FileNotFoundError as e:
        logger.debug(e)
        if os.environ.get("CONFIG_FOLDER_PATH"):
            logger.error(
                "Configuration file not found at {}.".format(
                    os.environ["CONFIG_FOLDER_PATH"]
                )
            )
            exit(0)
        else:
            logger.error(
                "No configuration file can be found. Please create a "
                "config.yml in one of these directories:\n"
                "{}".format(", ".join(CONFIG_DIRS))
            )
            exit(0)


def setup_logger(file=True, stderr=True):
    try:
        log_level = get_config()["config"].get("LOG_LEVEL", "INFO")
    except KeyError:
        log_level = "DEBUG"

    logger.remove()

    try:
        if stderr:
            logger.add(sys.stderr, level=log_level)

        if file:
            log_path = get_config().get("config", {}).get("LOG_DIR", None)

            if log_path and os.path.isdir(log_path):
                logger.add(
                    os.path.join(log_path, "output.log"),
                    rotation="5 MB",
                    level=log_level,
                )
            else:
                logger.remove()
                logger.add(sys.stderr, level="DEBUG")
                logger.error(
                    f"Logging directory {log_path} does not exists. Set a proper 'LOG_DIR' var in configuration file"
                )
                exit(0)

    except Exception as e:
        logger.remove()
        logger.add(sys.stderr, level="DEBUG")
        logger.exception("Something went wrong", e)
        exit(0)


class InterceptHandler(logging.Handler):
    def emit(self, record):
        # Skip if this is a loguru internal message
        if record.name.startswith('loguru.'): 
            return

        # Get logger name, default to root if none
        logger_name = record.name if record.name else 'root'
        
        # Map logging levels to Loguru levels
        level = logger.level(record.levelname).name

        # Extract log metadata
        frame, depth = logging.currentframe(), 2
        while frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1

        # Preserve exception info
        exc_info = record.exc_info if record.exc_info else None
        
        # Forward to loguru with original context
        logger.opt(depth=depth, exception=exc_info) \
              .bind(logger_name=logger_name) \
              .log(level, record.getMessage())

# Configure loguru
logger.configure(
    handlers=[
        {
            "sink": sys.stdout,
            "format": "<green>{time:YYYY-MM-DD HH:mm:ss}</green> | <level>{level: <8}</level> | <cyan>{extra[logger_name]}</cyan> | <level>{message}</level>"
        }
    ]
)

# Intercept all standard library logging
logging.basicConfig(handlers=[InterceptHandler()], level=0, force=True)

# Remove default handlers to avoid duplicates
for name in logging.root.manager.loggerDict.keys():
    logging.getLogger(name).handlers = []


import napalm
from loguru import logger
from napalm.pyIOSXR.exceptions import XMLCLIError

from netbox_network_importer.config import get_config

driver_iosxr = napalm.get_network_driver('iosxr')
driver_ios = napalm.get_network_driver('ios')
driver_iosxr_netconf = napalm.get_network_driver('iosxr_netconf')
driver_nxos = napalm.get_network_driver('nxos_ssh')
driver_junos = napalm.get_network_driver('junos')


class Napalm:
    def __init__(self, address, driver):
        self.address = address
        self.driver = driver

        # connection to napalm device
        # supports https://napalm.readthedocs.io/en/latest/base.html commands
        self.connection = None

    # set napalm_driver
    def __set_napalm_driver(self, hostname, driver, port=22):
        conn_param = {
            "hostname": hostname,
            "username": get_config()['tacacs']['TACACS_USERNAME'],
            "password": get_config()['tacacs']['TACACS_PASSWORD'],
            "timeout": 60,
            "optional_args": {
                "fast_cli": False,
                "allow_agent": False,
                "look_for_keys": False,
                "conn_timeout": 30,
                "secret": get_config()['tacacs']['TACACS_PASSWORD']
            }
        }

        if driver == "ios" or driver == "iosxe":
            return driver_ios(**conn_param)
        elif driver == "iosxr":
            return driver_iosxr(**conn_param)
        elif driver == "iosxr_netconf":
            conn_param['optional_args'] = {}
            conn_param['optional_args']['port'] = port
            return driver_iosxr_netconf(**conn_param)
        elif driver == 'nxos' or driver == 'nxos_ssh':
            conn_param['optional_args'] = {'read_timeout_override': 300}
            return driver_nxos(**conn_param)
        elif driver == 'junos':
            return driver_junos(**conn_param)
        else:
            logger.critical(
                "Cannot connect do this type of device: %s", driver)
            return False

    def open(self):
        try:
            self.connection = self.__set_napalm_driver(
                self.address, self.driver)
            self.connection.open()

        except Exception as e:
            #logger.critical("Unable to Connect, Global exception")
            raise e

    def close(self):
        self.connection.close()
        self.connection = None

    def get_interfaces_ip(self):
        try:
            self.open()
            res = self.connection.get_interfaces_ip()
            self.close()
            return res
        except XMLCLIError as e:
            if self.driver == 'iosxr':
                self.driver = 'iosxr_netconf'
                return self.get_interfaces_ip()
        except Exception as e:
            raise e

    def get_interfaces(self):
        try:
            self.open()
            res = self.connection.get_interfaces()
            self.close()
            return res
        except XMLCLIError as e:
            if self.driver == 'iosxr':
                self.driver = 'iosxr_netconf'
                return self.get_interfaces()
        except Exception as e:
            raise e

    def get_facts(self):
        try:
            self.open()
            res = self.connection.get_facts()
            self.close()
            return res
        except XMLCLIError as e:
            if self.driver == 'iosxr':
                self.driver = 'iosxr_netconf'
                return self.get_facts()
        except Exception as e:
            raise e

    def connection_test(self):
        try:
            self.open()
            # self.connection.get_facts() # Throws error for some IOSXR devices, Opened issue on Napalm
            self.connection.get_interfaces()
            self.close()
            return (True, f"{self.address} connected successfuly")
        except XMLCLIError as e:
            if self.driver == 'iosxr':
                self.driver = 'iosxr_netconf'
                return self.connection_test()
            else:
                return (False, e)
        except Exception as e:
            return (False, e)



from nornir import InitNornir

from netbox_network_importer.config import get_config


class Nornir:
    def init_nornir_w_netbox(
        filter_parameters={}, logging_enabled=False, logging_level="DEBUG"
    ):
        """Init nornir with Netbox.

        :param filter_parameters: params to filter certain devices
        """

        nr = InitNornir(
            logging={"enabled": logging_enabled, "level": logging_level},
            inventory={
                "plugin": "NetBoxInventory2",
                "options": {
                    "nb_url": get_config()["netbox"]["NETBOX_INSTANCE_URL"],
                    "nb_token": get_config()["netbox"]["NETBOX_API_TOKEN"],
                    "flatten_custom_fields": False,
                    "filter_parameters": filter_parameters,
                    "use_platform_slug": False,
                    "use_platform_napalm_driver": False,
                    # group_file,
                    # defaults_file - could be used for username and password for devices - connection options
                },
            },
            runner={
                "plugin": "threaded",
                "options": {
                    "num_workers": 20  # default
                },
            },
        )

        return nr



from enum import Enum


class Status(Enum):
    CHANGED = 1
    NOT_CHANGED = 2
    FAILED = 3
    ANOMALLY = 4
    CHECK_MANUALLY = 5
    EXCEPTION = 6
    SKIPPED = 7


class Action(Enum):
    CREATE = 1
    UPDATE = 2
    DELETE = 3
    LOOKUP = 4
    UNKNOWN = 5


class NetboxResult():
    result: str
    status: Status
    diff: dict
    action: Action
    exception: Exception

    def __init__(self, result: str, status: Status, action: Action, diff: dict, exception=None):
        self.result = result
        self.status = status
        self.diff = diff
        self.action = action
        self.exception = exception

    def to_dict(self):
        return {
            "result": self.result,
            "status": self.status.name,
            "action": self.action.name,
            "exception": self.exception,
            "diff": self.diff,
        }

    def __repr__(self):
        return self.result


class TaskNetboxResult():
    """Representation of results on Netbox though task

    :param netbox_results: list of NetboxResult
    :param task: task name
    """

    netbox_results = []  # type NetboxResult
    task = str  # task name

    def __init__(self, task: str, netbox_results: list, status, exception=None, comment=None):
        self.netbox_results = netbox_results
        self.task = task
        self.status = status
        self.exception = exception
        self.comment = comment

    def to_dict(self, except_status_codes=[]):
        """Returns dict representation of results of task action
        {
            '<task_name>': {                                                                                                      
                '<status_code>': [                                                                                                             
                    {NetboxResult.to_dict(),
                ]
        }

        :param netbox_results: list of NetboxResult
        :param except_status_codes: Does not return certain status codes
        :param task: task name
        """

        if except_status_codes:
            except_status_codes = [except_status_codes] if not isinstance(
                except_status_codes, list) else except_status_codes

        result = {self.task: {"status": self.status,
                              "comment": self.comment,
                              "results": {},
                              "exception": self.exception, }
                  }

        for res in self.netbox_results:
            if res.status in except_status_codes:
                continue

            res_dict = res.to_dict()
            if not result[self.task]['results'].get(res_dict['status'], None):
                result[self.task]['results'][res_dict['status']] = []

            result[self.task]['results'][res.status.name].append(res_dict)

        return result

    def __repr__(self):
        return self.task



import re
from collections import defaultdict

from dictdiffer import diff
from netutils.interface import canonical_interface_name, canonical_interface_name_list
from netutils.vlan import vlanconfig_to_list

PYATS_PLATFORM_MAPPER = {
    'ios': "ios",
    'cisco_ios': "ios",
    'iosxr': "iosxr",
    'cisco_iosxr': "iosxr",
    'iosxe': 'iosxe',
    'cisco_iosxe': 'iosxe',
    'nxos': "nxos_ssh",
    'nxos_ssh': 'nxos_ssh',
    'junos': 'junos',
    'juniper': 'junos',
    'sros': 'sros',
    'nokia': 'sros'
}

BASE_INTERFACES_EXTENDED = {
    "TenGigE": "TenGigabitEthernet",
    "Ap": "AppGigabitEthernet"
}


def determine_pyats_platform(os):
    return PYATS_PLATFORM_MAPPER.get(os, os)


def is_interface_physical(name):  # pylint: disable=R0911
    """Function evaluate if an interface is likely to be a physical interface.
    Args:
      name (str): name of the interface to evaluate
    Return:
      True, False or None
    """
    # Match most physical interface Cisco that contains Ethernet
    #  GigabitEthernet0/0/2
    #  GigabitEthernet0/0/2:3
    #  TenGigabitEthernet0/0/4
    cisco_physical_intf = r"^[a-zA-Z]+[Ethernet][0-9\/\:]+$"

    # Match Sub interfaces finishing with ".<number>"
    sub_intf = r".*\.[0-9]+$"

    # Regex for loopback and vlan interface
    loopback = r"^(L|l)(oopback|o)[0-9]+$"
    vlan = r"^(V|v)(lan)[0-9]+$"

    # Generic physical interface match
    #  mainly looking for <int>/<int> or <int>/<int>/<int> at the end
    generic_physical_intf = r"^[a-zA-Z\-]+[0-9]+\/[0-9\/\:]+$"

    # Match Juniper Interfaces
    jnpr_physical_intf = r"^[a-z]+\-[0-9\/\:]+$"

    if re.match(loopback, name):
        return False
    if re.match(vlan, name):
        return False
    if re.match(cisco_physical_intf, name):
        return True
    if re.match(sub_intf, name):
        return False
    if re.match(jnpr_physical_intf, name):
        return True
    if re.match(generic_physical_intf, name):
        return True

    return None


def is_interface_lag(name):
    """Function to evaluate if an interface is likely to be a lag.
    Args:
      name (str): name of the interface to evaluate
    Return:
      True, False or None
    """
    port_channel_intf = r"^port\-channel[0-9]+$"
    po_intf = r"^po[0-9]+$"
    ae_intf = r"^ae[0-9]+$"
    bundle_intf = r"^Bundle\-Ether[0-9]+$"

    if re.match(port_channel_intf, name.lower()):
        return True
    if re.match(ae_intf, name):
        return True
    if re.match(po_intf, name):
        return True
    if re.match(bundle_intf, name):
        return True

    return None


def get_netbox_interface_type(ifc_name):
    is_physical = is_interface_physical(ifc_name)
    is_lag = is_interface_lag(ifc_name)

    if is_lag:
        return 'lag'
    elif is_physical is False:
        return 'virtual'
    else:
        return 'other'


def canonical_interface_name_edited(interface):
    return canonical_interface_name(interface, BASE_INTERFACES_EXTENDED)


def canonical_interface_name_list_edited(interfaces):
    return canonical_interface_name_list(interfaces, BASE_INTERFACES_EXTENDED)


def get_diff(before, after):
    diff_result = defaultdict(dict)
    changes = list(diff(before, after))
    for change in changes:
        type = change[0]
        name = change[1]
        diffs = change[2]

        if isinstance(name, list) and len(name) == 2:
            name = name[1]

        diff_result[type][name] = list(diffs)

    return dict(diff_result)


def get_address_from_netbox_task(task):
    return task.host.get("primary_ip", {}).get("address", "").split('/')[0]


def vlanconfig_to_list_edited(trunks_data):
    try:
        return vlanconfig_to_list(trunks_data)
    except ValueError as e:
        return []
    except Exception as e:
        raise e


def remove_key_from_results(results: dict, key="NOT_CHANGED"):
    results_without_not_changed_data = dict(results)
    # Remove NOT_CHANGED keys from results
    for device, tasks in results_without_not_changed_data.items():
        for task, output in tasks.items():
            if output.get("results", {}).get(key, None):
                output['results'].pop(key)

    return results_without_not_changed_data



from netbox_network_importer.connections.Netbox import Netbox
from netbox_network_importer.helper import (
    canonical_interface_name_edited,
    get_diff,
    get_netbox_interface_type,
)
from netbox_network_importer.results.results import Action, NetboxResult, Status


def process_lag_interfaces(host, parsed_lag_interfaces) -> list:
    """Update netbox lag interfaces
    - Update lag interfaces, which already exists in netbox

    :param parsed_interfaces: parsed interfaces for netbox operations
    """

    RESULTS = []

    # Init connection to Netbox via PyNetbox
    nb = Netbox.connect()

    # find the device instance in netbox
    dev = nb.dcim.devices.get(host.data['id'])

    # NETBOX LAG INTERFACES
    NB_LAG_INTERFACES = nb.dcim.interfaces.filter(device_id=dev.id, type='lag')
    NB_LAG_INTERFACES_dict = {ifc.name: ifc for ifc in [
        ifc for ifc in NB_LAG_INTERFACES]}

    # NETBOX LAG CHILDREN
    NB_LAG_CHILDREN = nb.dcim.interfaces.filter(
        device_id=dev.id, lag_id__n='null')
    NB_LAG_CHILDREN_dict = {ifc.name: ifc for ifc in [
        ifc for ifc in NB_LAG_CHILDREN]}

    for parent_lag, children_of_lag in parsed_lag_interfaces.items():
        nb_parent_lag_ifc = nb.dcim.interfaces.get(
            device_id=dev.id, name__ie=canonical_interface_name_edited(parent_lag))

        # if parent not found in netbox, skip to next record
        if not nb_parent_lag_ifc:
            RESULTS.append(
                NetboxResult(result=f"Parent LAG interface: {parent_lag} - could not be found in netbox on device: {host.name}", diff={
                }, action=Action.LOOKUP, status=Status.ANOMALLY)
            )
            continue

        # Skip IGNORED interfaces
        if nb_parent_lag_ifc.custom_fields.get("ignore_importer", False) == True:
            RESULTS.append(NetboxResult(
                result=f"{nb_parent_lag_ifc.name} - Ignored by Importer - Skipping", status=Status.SKIPPED, action=Action.LOOKUP, diff=""))
            NB_LAG_INTERFACES_dict.pop(nb_parent_lag_ifc.name, None)

        else:
            # SET parent interface as lag
            RESULTS.append(interface_parent_update(
                nb_ifc=nb_parent_lag_ifc, params={'type': 'lag'}))

            # Pop processed parent of already existing lags
            NB_LAG_INTERFACES_dict.pop(nb_parent_lag_ifc.name, None)

        for lag_child in children_of_lag:
            nb_child_ifc = nb.dcim.interfaces.get(
                device_id=dev.id, name__ie=canonical_interface_name_edited(lag_child))

            # if child not found in netbox, skip to next record
            if not nb_child_ifc:
                RESULTS.append(
                    NetboxResult(result=f"Child interface {canonical_interface_name_edited(lag_child)} could not be found in netbox on device {host.name}",
                                 action=Action.LOOKUP, status=Status.ANOMALLY, diff={})

                )
                continue

            # Skip IGNORED interfaces
            if nb_child_ifc.custom_fields.get("ignore_importer", False) == True:
                RESULTS.append(NetboxResult(
                    result=f"{nb_child_ifc.name} - Ignored by Importer - Skipping", status=Status.SKIPPED, action=Action.LOOKUP, diff=""))
                NB_LAG_CHILDREN_dict.pop(nb_child_ifc.name, None)
                continue
            elif nb_parent_lag_ifc.custom_fields.get("ignore_importer", False) == True:
                RESULTS.append(NetboxResult(
                    result=f"{nb_child_ifc.name} - PARENT Ignored by Importer - Skipping", status=Status.SKIPPED, action=Action.LOOKUP, diff=""))
                NB_LAG_CHILDREN_dict.pop(nb_child_ifc.name, None)
                continue
            else:
                # set parent to a child interface
                RESULTS.append(interface_child_update(
                    nb_ifc=nb_child_ifc, params={'lag': nb_parent_lag_ifc}))

                # Pop processed child of alredy existing children
                NB_LAG_CHILDREN_dict.pop(nb_child_ifc.name, None)

    # TODO: Process (remove) lags which are not parsed
    # unset parent
    for ifc_name, nb_ifc in NB_LAG_CHILDREN_dict.items():
        RESULTS.append(interface_child_update(
            nb_ifc=nb_ifc, params={'lag': None}))

    # change type from lag to virtual
    # TODO: Lag without children can exists
    for ifc_name, nb_ifc in NB_LAG_INTERFACES_dict.items():
        RESULTS.append(interface_parent_update(
            nb_ifc=nb_ifc, params={'type': get_netbox_interface_type(ifc_name)}))

    return RESULTS


def interface_parent_update(nb_ifc, params):
    before = get_changes_parent_interface(nb_ifc)

    if nb_ifc.update(params):
        # TODO: Reload ifc
        nb_ifc = nb_ifc.api.dcim.interfaces.get(nb_ifc.id)
        after = get_changes_parent_interface(nb_ifc)

        return NetboxResult(
            result=f"{nb_ifc.name} - LAG updated successfully",
            status=Status.CHANGED,
            action=Action.UPDATE,
            diff=get_diff(before, after)
        )
    else:
        return NetboxResult(
            result=f"{nb_ifc.name} - LAG - nothing to update",
            status=Status.NOT_CHANGED,
            action=Action.UPDATE,
            diff={}
        )


def interface_child_update(nb_ifc, params):
    before = get_changes_child_interface(nb_ifc)

    if nb_ifc.update(params):
        # TODO: Reload ifc
        nb_ifc = nb_ifc.api.dcim.interfaces.get(nb_ifc.id)
        after = get_changes_child_interface(nb_ifc)

        return NetboxResult(
            result=f"{nb_ifc.name} - LAG - Parent set successfully",
            status=Status.CHANGED,
            action=Action.UPDATE,
            diff=get_diff(before, after)
        )
    else:
        return NetboxResult(
            result=f"{nb_ifc.name} - nothing to update",
            status=Status.NOT_CHANGED,
            action=Action.UPDATE,
            diff={}
        )


def get_changes_parent_interface(nb_ifc):
    return {'type': nb_ifc.type.label}


def get_changes_child_interface(nb_ifc):
    if nb_ifc.lag:
        return {'lag': nb_ifc.lag.name}
    else:
        return {'lag': None}



from nornir.core.task import Task, Result
from netbox_network_importer.results.results import TaskNetboxResult, Status
from netbox_network_importer.connections.Napalm import Napalm
from netbox_network_importer.helper import get_address_from_netbox_task


def connection_test(task: Task, pyats) -> Result:
    if not task.host.platform:
        task_result = TaskNetboxResult(task=task.name,
                                       netbox_results=[],
                                       status="failed",
                                       exception="Platform is not set",
                                       comment="Platform is not set")
        return Result(
            host=task.host,
            exception=f"{task.host.name} - Platform is not set",
            result=f"{task.host.name} - Platform is not set",
            failed=True,
            netbox_results=task_result,
            diff=task_result.to_dict())

    elif task.host.platform not in ['ios', 'iosxr', 'iosxe', 'nxos', 'nxos_ssh', 'junos']:
        task_result = TaskNetboxResult(task=task.name,
                                       netbox_results=[],
                                       status="failed",
                                       exception=f"Unsupported platform {task.host.platform}",
                                       comment=f"Unsupported platform {task.host.platform}")
        return Result(
            host=task.host,
            exception=f"{task.host.name} - Unsupported platform {task.host.platform}",
            result=f"{task.host.name} - Unsupported platform {task.host.platform}",
            failed=True,
            netbox_results=task_result,
            diff=task_result.to_dict())

    # NAPALM connection test
    host_ip = get_address_from_netbox_task(task)

    if not host_ip:
        task_result = TaskNetboxResult(task=task.name,
                                       netbox_results=[],
                                       status="failed",
                                       exception=f"Primary IP not set",
                                       comment=f"Primary IP not set")
        return Result(
            host=task.host,
            exception=f"{task.host.name} - Primary IP not set",
            result=f"{task.host.name} - Primary IP not set",
            failed=True,
            netbox_results=task_result,
            diff=task_result.to_dict())

    dev = Napalm(host_ip, task.host.platform)
    connected, message = dev.connection_test()
    if not connected:
        task_result = TaskNetboxResult(task=task.name,
                                       netbox_results=[],
                                       status="failed",
                                       exception=message,
                                       comment="Failed to connect to device via Napalm")

        return Result(
            host=task.host,
            exception=message,
            result=f"{task.host.name} - Failed to connect to napalm",
            failed=True,
            netbox_results=task_result,
            diff=task_result.to_dict(except_status_codes=[Status.NOT_CHANGED]))

    # Pyats Connection Test
    connected, message = pyats.connection_test(hostname=task.host.name)
    if not connected:
        task_result = TaskNetboxResult(task=task.name,
                                       netbox_results=[],
                                       status="failed",
                                       exception=message,
                                       comment="Unable to connect to device via PyATS")
        return Result(
            host=task.host,
            exception=message,
            result=f"{task.host.name} - Unable to connect to device",
            netbox_results=task_result,
            diff=task_result.to_dict(except_status_codes=[Status.NOT_CHANGED]),
            failed=True)

    # Return OK

    task_result = TaskNetboxResult(task=task.name,
                                   netbox_results=[],
                                   status="completed")
    return Result(host=task.host,
                  result="Connection test - OK",
                  netbox_results=task_result,
                  diff=task_result.to_dict(except_status_codes=[Status.NOT_CHANGED]))



import datetime
from netbox_network_importer.connections.Netbox import Netbox
from netbox_network_importer.results.results import NetboxResult, Status, Action
from netbox_network_importer.helper import get_diff


def process_facts(host, napalm_facts) -> list:
    RESULTS = []  # list of NetboxResult
    # Init connection to Netbox via PyNetbox
    nb = Netbox.connect()
    # find the device instance in netbox
    dev = nb.dcim.devices.get(host.data['id'])
    # store params before changes
    before = get_changed_params(dev)

    dev.serial = napalm_facts.get("serial_number", "")
    dev.custom_fields['ni_update'] = True
    dev.custom_fields['ni_update_date'] = str(datetime.date.today())

    if dev.save():
        # store params after changes
        after = get_changed_params(dev)
        RESULTS.append(
            NetboxResult(
                result=f"{dev.name} - Serial number updated: {dev.serial}",
                status=Status.CHANGED,
                action=Action.UPDATE,
                diff=get_diff(before, after)
            )
        )
    else:
        RESULTS.append(
            NetboxResult(
                result=f"{dev.name} - Serial number already exists: {dev.serial}",
                status=Status.NOT_CHANGED,
                action=Action.UPDATE,
                diff={}
            )
        )
    return RESULTS


def get_changed_params(nb_device):
    return {"serial": nb_device.serial,
            "ni_update": nb_device.custom_fields['ni_update'],
            "ni_update_date": nb_device.custom_fields['ni_update_date']}



# note that these imports are only needed if you are annotating your code with types
from typing import Dict

from nornir.core.inventory import Host
from nornir.core.task import AggregatedResult, MultiResult, Task


class SaveNetboxResults:
    def __init__(self, data: Dict[str, None]) -> None:
        self.data = data

    def task_started(self, task: Task) -> None:
        pass

    def task_completed(self, task: Task, result: AggregatedResult) -> None:
        pass

    def task_instance_started(self, task: Task, host: Host) -> None:
        pass

    def task_instance_completed(
        self, task: Task, host: Host, result: MultiResult
    ) -> None:
        if not self.data.get(host.name, None):
            self.data[host.name] = {}

        if hasattr(result, "netbox_results"):
            self.data[host.name].update(result.netbox_results.to_dict())

    def subtask_instance_started(self, task: Task, host: Host) -> None:
        pass

    def subtask_instance_completed(
        self, task: Task, host: Host, result: MultiResult
    ) -> None:
        pass



import genie


def show_interfaces(pyats, hostname):
    """Get `show interfaces` command parsed by Genie.

    :param pyats: passed pyats connection instance with loaded testbed
    """
    try:
        dev = pyats.connect_device(hostname=hostname)
        return dev.parse("show interfaces")
    except Exception as e:
        raise e


def show_interface(pyats, hostname):
    """Get `show interfaces` command parsed by Genie.

    :param pyats: passed pyats connection instance with loaded testbed
    """
    dev = pyats.connect_device(hostname=hostname)
    show_ifcs = dev.parse("show interface")

    return show_ifcs


def show_interface_switchport(pyats, hostname):
    """Get `show interfaces` command parsed by Genie.

    :param pyats: passed pyats connection instance with loaded testbed
    """
    dev = pyats.connect_device(hostname=hostname)
    interface_switchport = dev.parse("show interface switchport")

    return interface_switchport


def show_vlan(pyats, hostname):
    try:
        dev = pyats.connect_device(hostname=hostname)
        return dev.parse("show vlan")
    except genie.metaparser.util.exceptions.InvalidCommandError:
        return {}
    except genie.metaparser.util.exceptions.SchemaEmptyParserError:
        return {}


def show_interfaces_status(pyats, hostname):
    try:
        dev = pyats.connect_device(hostname=hostname)
        return dev.parse("show interfaces status")
    except genie.metaparser.util.exceptions.InvalidCommandError:
        return {}
    except genie.metaparser.util.exceptions.SchemaEmptyParserError:
        return {}


def show_interfaces_trunk(pyats, hostname):
    try:
        dev = pyats.connect_device(hostname=hostname)
        return dev.parse("show interfaces trunk")
    except genie.metaparser.util.exceptions.InvalidCommandError:
        return {}
    except genie.metaparser.util.exceptions.SchemaEmptyParserError:
        return {}


def learn_interfaces(pyats, hostname):
    """PyATS Learn Device Features `learn interface` command parsed by pyATS.

    :param pyats: passed pyats connection instance with loaded testbed
    """

    dev = pyats.connect_device(hostname=hostname)

    learn_ifcs = dev.learn("interface")
    return learn_ifcs.info



import jsonpickle
from json2html import *
from loguru import logger
import time
from netbox_network_importer.config import get_config

TIMESTAMP = time.strftime("%Y%m%d-%H%M%S")


def json_output(results: dict):
    try:
        OUTPUT_DIR = get_config()['config'].get('OUTPUT_DIR')
        json_file_path = f"{OUTPUT_DIR}/{TIMESTAMP}-result.json"
        json_data = jsonpickle.encode(
            results, keys=True, max_depth=20, indent=4)
        with open(json_file_path, "w") as outfile:
            outfile.write(json_data)
    except Exception as e:
        logger.error("Unable to results as json")
        logger.error(e)


def html_output(results: dict):
    try:
        OUTPUT_DIR = get_config()['config'].get('OUTPUT_DIR')
        html_file_path = f"{OUTPUT_DIR}/{TIMESTAMP}-result.html"
        html_json_data = jsonpickle.encode(results, keys=True, max_depth=20)
        html_data = json2html.convert(json=html_json_data)
        with open(html_file_path, "w") as outfile:
            outfile.write(html_data)
    except Exception as e:
        logger.error("Unable to results as html")
        logger.error(e)



from nornir.core.task import Result, Task

from netbox_network_importer.connections.Napalm import Napalm
from netbox_network_importer.data_converter.network_to_netbox import (
    convert_bandwidth_interfaces,
    convert_interfaces,
    convert_interfaces_ips,
    convert_interfaces_vlans_iosxe,
    convert_interfaces_vlans_nxos,
    convert_lag_interfaces,
)
from netbox_network_importer.helper import (
    get_address_from_netbox_task,
    is_interface_lag,
)
from netbox_network_importer.netbox_updater.bandwidth import (
    process_bandwidth_interfaces,
)
from netbox_network_importer.netbox_updater.facts import process_facts
from netbox_network_importer.netbox_updater.interfaces import process_interfaces
from netbox_network_importer.netbox_updater.interfaces_ips import process_interfaces_ips
from netbox_network_importer.netbox_updater.interfaces_vlans import (
    process_interfaces_vlans,
)
from netbox_network_importer.netbox_updater.lag_interfaces import process_lag_interfaces
from netbox_network_importer.pyats.parser import (
    show_interface,
    show_interface_switchport,
    show_interfaces,
    show_interfaces_status,
    show_interfaces_trunk,
    show_vlan,
)
from netbox_network_importer.results.results import Status, TaskNetboxResult


def synchronize_serials(task: Task, pyats) -> Result:
    # Parse interfaces information from network
    try:
        host_ip = get_address_from_netbox_task(task)
        dev = Napalm(host_ip, task.host.platform)
        napalm_facts = dev.get_facts()
    except Exception as e:
        task_result = TaskNetboxResult(
            task=task.name,
            netbox_results=[],
            status="failed",
            exception=e,
            comment="Failed to get facts data",
        )
        return Result(
            host=task.host,
            exception=e,
            result=f"{task.host.name} - Failed to get facts data via Napalm",
            failed=True,
            netbox_results=task_result,
            diff=task_result.to_dict(except_status_codes=[Status.NOT_CHANGED]),
        )

    # CRUD operations on Netbox with received data
    nb_results_list = process_facts(host=task.host, napalm_facts=napalm_facts)

    task_result = TaskNetboxResult(
        task=task.name, netbox_results=nb_results_list, status="completed"
    )
    return Result(
        host=task.host,
        result="Facts were processed",
        netbox_results=task_result,
        diff=task_result.to_dict(except_status_codes=[Status.NOT_CHANGED]),
    )


def synchronize_interfaces(task: Task, pyats) -> Result:
    """Sync interfaces of device between network and Netbox.

    List of tasks that poll data from the network (via PyAts with genie parser),
    process it into the required form (converts data to netbox friendly params),
    executes the CRUD action on Netbox

    :param pyats: passed pyats connection instance with loaded testbed
    """

    # Parse interfaces information from network
    try:
        host_ip = get_address_from_netbox_task(task)
        dev = Napalm(host_ip, task.host.platform)
        napalm_interfaces = dev.get_interfaces()
    except Exception as e:
        task_result = TaskNetboxResult(
            task=task.name,
            netbox_results=[],
            status="failed",
            exception=e,
            comment="Failed to get interfaces data",
        )
        return Result(
            host=task.host,
            exception=e,
            result=f"{task.host.name} - Failed to get interfaces data via Napalm",
            failed=True,
            netbox_results=task_result,
            diff=task_result.to_dict(except_status_codes=[Status.NOT_CHANGED]),
        )

    # Convert parsed information to Netbox friendly data (fields)
    converted_ifcs = convert_interfaces(napalm_interfaces=napalm_interfaces)

    # CRUD operations on Netbox with received data
    nb_results_list = process_interfaces(
        host=task.host, parsed_interfaces=converted_ifcs
    )

    task_result = TaskNetboxResult(
        task=task.name, netbox_results=nb_results_list, status="completed"
    )
    return Result(
        host=task.host,
        result="Interfaces were processed",
        netbox_results=task_result,
        diff=task_result.to_dict(except_status_codes=[Status.NOT_CHANGED]),
    )


def synchronize_interfaces_lags(task: Task, pyats) -> Result:
    """Sync lags of device between network and Netbox.

    List of tasks that poll data from the network (via PyAts with genie parser),
    process it into the required form (converts data to netbox friendly params),
    executes the CRUD action on Netbox

    :param pyats: passed pyats connection instance with loaded testbed
    """

    try:
        if task.host.platform in ["nxos", "nxos_ssh"]:
            genie_interfaces = show_interface(pyats=pyats, hostname=task.host.name)
        else:
            genie_interfaces = show_interfaces(pyats=pyats, hostname=task.host.name)
    except Exception as e:
        task_result = TaskNetboxResult(
            task=task.name,
            netbox_results=[],
            status="failed",
            exception=e,
            comment="Unable to parse show_interfaces via PyATS",
        )
        return Result(
            host=task.host,
            exception=e,
            result=f"{task.host.name} - Unable to parse show_interfaces via PyATS",
            netbox_results=task_result,
            diff=task_result.to_dict(except_status_codes=[Status.NOT_CHANGED]),
            failed=True,
        )

    # Filter lag interfaces
    genie_lag_interfaces = {
        key: value for key, value in genie_interfaces.items() if is_interface_lag(key)
    }

    # Convert parsed information to Netbox friendly data (fields)
    converted_lag_ifcs = convert_lag_interfaces(
        genie_lag_interfaces=genie_lag_interfaces, platform=task.host.platform
    )
    # CRUD operations on Netbox with received data
    nb_results_list = process_lag_interfaces(
        host=task.host, parsed_lag_interfaces=converted_lag_ifcs
    )

    status = "no_data_parsed" if not converted_lag_ifcs else "completed"
    result = "No lags were parsed" if not converted_lag_ifcs else "LAGS were processed"

    task_result = TaskNetboxResult(task.name, nb_results_list, status=status)

    return Result(
        host=task.host,
        result=result,
        netbox_results=task_result,
        diff=task_result.to_dict(except_status_codes=[Status.NOT_CHANGED]),
    )


def synchronize_interfaces_ips(task: Task, pyats) -> Result:
    """Sync interfaces IP addresses of device between network and Netbox.

    List of tasks that poll data from the network (via PyAts with genie parser),
    process it into the required form (converts data to netbox friendly params),
    executes the CRUD action on Netbox
    """

    try:
        host_ip = get_address_from_netbox_task(task)
        dev = Napalm(host_ip, task.host.platform)
        napalm_interfaces_ip = dev.get_interfaces_ip()
    except Exception as e:
        task_result = TaskNetboxResult(
            task=task.name,
            netbox_results=[],
            status="failed",
            exception=e,
            comment="Failed to get interfaces ip info Napalm",
        )
        return Result(
            host=task.host,
            exception=e,
            result=f"{task.host.name} - Failed to get interfaces ip info Napalm",
            failed=True,
            netbox_results=task_result,
            diff=task_result.to_dict(except_status_codes=[Status.NOT_CHANGED]),
        )

    # parse ips them
    converted_ifcs = convert_interfaces_ips(
        napalm_interfaces_ip=napalm_interfaces_ip, platform=task.host.platform
    )

    # CRUD operations on Netbox with received data
    nb_results_list = process_interfaces_ips(
        host=task.host, parsed_interfaces=converted_ifcs
    )

    task_result = TaskNetboxResult(
        task=task.name, netbox_results=nb_results_list, status="completed"
    )

    return Result(
        host=task.host,
        result="IPs were processed",
        netbox_results=task_result,
        diff=task_result.to_dict(except_status_codes=[Status.NOT_CHANGED]),
        status="completed",
    )


def synchronize_interfaces_vlans(task: Task, pyats) -> Result:
    # Parse interfaces information from network

    try:
        if task.host.platform == "iosxr":
            task_result = TaskNetboxResult(task.name, [], status="IOSXR - SKIPPED")
            return Result(
                host=task.host,
                result=f"{task.host.name} {task.host.platform} - Skipping IOSXR vlans synchronization",
                failed=False,
                netbox_results=task_result,
                diff={},
            )

        elif task.host.platform == "iosxe" or task.host.platform == "ios":
            vlans = show_vlan(pyats=pyats, hostname=task.host.name)
            vlans_interfaces = show_interfaces_status(
                pyats=pyats, hostname=task.host.name
            )
            trunks = show_interfaces_trunk(pyats=pyats, hostname=task.host.name)

            RESULT = convert_interfaces_vlans_iosxe(vlans, vlans_interfaces, trunks)
        elif task.host.platform in ["nxos", "nxos_ssh"]:
            interface_switchport = show_interface_switchport(
                pyats=pyats, hostname=task.host.name
            )
            vlan = show_vlan(pyats=pyats, hostname=task.host.name)
            RESULT = convert_interfaces_vlans_nxos(interface_switchport, vlan)
        else:
            raise NotImplementedError
    except Exception as e:
        task_result = TaskNetboxResult(
            task=task.name,
            netbox_results=[],
            status="failed",
            exception=e,
            comment=f"{task.host.name} {task.host.platform}- Failed to get/parse vlan information",
        )
        return Result(
            host=task.host,
            exception=e,
            result=f"{task.host.name} {task.host.platform}- Failed to get/parse vlan information",
            failed=True,
            netbox_results=task_result,
            diff=task_result.to_dict(except_status_codes=[Status.NOT_CHANGED]),
        )

    # CRUD operations on Netbox with received data
    nb_results_list = process_interfaces_vlans(host=task.host, parsed_data=RESULT)

    task_result = TaskNetboxResult(task.name, nb_results_list, status="completed")

    # TODO:
    failed = False
    if task_result.to_dict().get(task.name).get("results", {}).get("EXCEPTION"):
        failed = True
        task_result.status = "FAILED"

    changed = False
    if task_result.to_dict().get(task.name).get("results", {}).get("CHANGED"):
        changed = True

    diff = task_result.to_dict(except_status_codes=[Status.NOT_CHANGED])
    return Result(
        host=task.host,
        result="VLANS were processed",
        netbox_results=task_result,
        diff=diff,
        failed=failed,
        changed=changed,
    )


def synchronize_interfaces_bandwidths(task: Task, pyats) -> Result:
    try:
        if task.host.platform in ["nxos", "nxos_ssh"]:
            genie_interfaces = show_interface(pyats=pyats, hostname=task.host.name)
        elif task.host.platform in ["iosxe", "ios", "iosxr"]:
            genie_interfaces = show_interfaces(pyats=pyats, hostname=task.host.name)
        else:
            task_result = TaskNetboxResult(
                task=task.name,
                netbox_results=[],
                status="failed",
                exception=NotImplementedError,
                comment=f"{task.host.name} {task.host.platform}- Not implemented",
            )
            return Result(
                host=task.host,
                exception=NotImplementedError,
                result=f"{task.host.name} {task.host.platform}- Not implemented",
                failed=True,
                netbox_results=task_result,
                diff=task_result.to_dict(except_status_codes=[Status.NOT_CHANGED]),
            )
    except Exception as e:
        task_result = TaskNetboxResult(
            task=task.name,
            netbox_results=[],
            status="failed",
            exception=e,
            comment="Unable to parse show_interfaces via PyATS",
        )
        return Result(
            host=task.host,
            exception=e,
            result=f"{task.host.name} - Unable to parse show_interfaces via PyATS",
            netbox_results=task_result,
            diff=task_result.to_dict(except_status_codes=[Status.NOT_CHANGED]),
            failed=True,
        )

    # Filter lag interfaces

    converted_bandwidth_ifcs = convert_bandwidth_interfaces(
        genie_interfaces=genie_interfaces, platform=task.host.platform
    )

    # CRUD operations on Netbox with received data
    nb_results_list = process_bandwidth_interfaces(
        host=task.host, converted_bandwidth_ifcs=converted_bandwidth_ifcs
    )

    status = "completed"
    result = "Bandwidth were processed"

    task_result = TaskNetboxResult(task.name, nb_results_list, status=status)

    return Result(
        host=task.host,
        result=result,
        netbox_results=task_result,
        diff=task_result.to_dict(except_status_codes=[Status.NOT_CHANGED]),
    )



from netbox_network_importer.connections.Netbox import Netbox
from netbox_network_importer.helper import (
    canonical_interface_name_edited,
    get_diff,
)
from netbox_network_importer.results.results import Action, NetboxResult, Status


def process_bandwidth_interfaces(host, converted_bandwidth_ifcs) -> list:
    RESULTS = []  # list of NetboxResult

    # Init connection to Netbox via PyNetbox
    nb = Netbox.connect()

    # find the device instance in netbox
    dev = nb.dcim.devices.get(host.data["id"])

    # find interfaces linked to the device
    ifcs_filter = nb.dcim.interfaces.filter(device_id=dev.id)

    # convert filtered interfaces into dictionary of pynetbox interface instances
    nb_interfaces_dict = {
        canonical_interface_name_edited(ifc.name): ifc
        for ifc in [ifc for ifc in ifcs_filter]
    }

    # setup NB parameters
    for ifc, bandwidth in converted_bandwidth_ifcs.items():
        # get NB interface instance
        nb_ifc = nb_interfaces_dict.get(ifc, None)

        if nb_ifc:
            # Skip IGNORED interfaces
            if nb_ifc.custom_fields.get("ignore_importer", False) == True:
                RESULTS.append(
                    NetboxResult(
                        result=f"{nb_ifc.name} - Ignored by Importer - Skipping",
                        status=Status.SKIPPED,
                        action=Action.LOOKUP,
                        diff="",
                    )
                )

                nb_interfaces_dict.pop(nb_ifc.name, None)
                continue
            else:
                # If interface already exists in netbox, update it
                RESULTS.append(
                    interface_bandwidth_update(
                        nb_ifc=nb_ifc, bandwidth=bandwidth)
                )
        else:
            RESULTS.append(
                NetboxResult(
                    result=f"Interface {ifc} not found in Netbox",
                    status=Status.ERROR,
                    action=Action.LOOKUP,
                    diff={},
                )
            )
    return RESULTS


def interface_bandwidth_update(nb_ifc, bandwidth):
    try:
        # Get data before changes
        before = nb_ifc.custom_fields["bandwidth"]

        if nb_ifc.update({"custom_fields": {"bandwidth": bandwidth}}):
            nb_ifc = nb_ifc.api.dcim.interfaces.get(nb_ifc.id)
            after = nb_ifc.custom_fields["bandwidth"]

            return NetboxResult(
                result=f"{nb_ifc.name} - saved successfully",
                status=Status.CHANGED,
                action=Action.UPDATE,
                diff=get_diff(before, after),
            )
        else:
            return NetboxResult(
                result=f"{nb_ifc.name} - nothing to do",
                action=Action.UPDATE,
                status=Status.NOT_CHANGED,
                diff={},
            )
    except Exception as e:
        return NetboxResult(
            result=f"{nb_ifc.name} - Exception Occurs: {e}",
            status=Status.EXCEPTION,
            action=Action.DELETE,
            diff={},
            exception=e,
        )



from netutils.mac import is_valid_mac, mac_to_format

from netbox_network_importer.helper import (
    canonical_interface_name_edited,
    canonical_interface_name_list_edited,
    get_netbox_interface_type,
    vlanconfig_to_list_edited,
)


def convert_interfaces(napalm_interfaces):
    """Convert Genie parsed interfaces to Netbox friendly fields.

    :param genie_interfaces: dictionary of genie parsed interfaces by `show interfaces` command
    :param platform: OS of host device (different OS return returns different genie parsed json)
    """

    converted_result = {}
    """Example of Netbox friendly format
    
    {
        "canonical_interface_name": {
            'enabled': True / False,
            'type': virtual / lag / other,
            'mtu': 1500 / None,
            'description': "Some description" / "",
            'mac_address': "MacAddress" if any
        }
    }
    """

    for name, properties in napalm_interfaces.items():
        canonical_name = canonical_interface_name_edited(name)
        converted_result[canonical_name] = {
            "enabled": properties["is_enabled"],
            "type": get_netbox_interface_type(canonical_name),
            "mtu": properties.get("mtu", None),
            "description": properties.get("description", ""),
        }

        if converted_result[canonical_name]["mtu"] < 1:
            converted_result[canonical_name]["mtu"] = None

        if is_valid_mac(properties.get("mac_address") or ""):
            converted_result[canonical_name]["mac_address"] = mac_to_format(
                properties.get("mac_address"), "MAC_COLON_TWO"
            ).upper()
        else:
            converted_result[canonical_name]["mac_address"] = None

    return converted_result


def convert_lag_interfaces(genie_lag_interfaces, platform="iosxr"):
    """Convert Genie parsed interfaces (lags only) to Netbox friendly fields.

    :param genie_lag_interfaces: dictionary of genie parsed interfaces (lags only) by `show interfaces` command
    :param platform: OS of host device (different OS return returns different genie parsed json)
    """

    converted_result = {}
    """Example of Netbox friendly format
    
    {
        'canonical_interface_lag_name':  ['Etherner1/1', 'Ethernet1/2'], #childrens in canonical names
        'canonical_interface_lag_name' ['Ethernet2/1'] #childrens in canonical names
    }
    """

    if platform == "iosxr":
        for parent_lag, properties in genie_lag_interfaces.items():
            if properties.get("port_channel"):
                if properties.get("port_channel").get("member_count"):
                    if properties.get("port_channel").get("member_count") > 0:
                        converted_result[
                            canonical_interface_name_edited(parent_lag)
                        ] = canonical_interface_name_list_edited(
                            properties.get("port_channel").get(
                                "members").keys()
                        )

    elif platform == "iosxe" or platform == "ios":
        for parent_lag, properties in genie_lag_interfaces.items():
            if properties.get("port_channel"):
                if properties.get("port_channel").get("port_channel_member_intfs"):
                    converted_result[canonical_interface_name_edited(parent_lag)] = (
                        canonical_interface_name_list_edited(
                            properties.get("port_channel").get(
                                "port_channel_member_intfs"
                            )
                        )
                    )

    elif platform == "nxos" or platform == "nxos_ssh":
        for parent_lag, properties in genie_lag_interfaces.items():
            if properties.get("port_channel"):
                if properties.get("port_channel").get("port_channel_member_intfs"):
                    converted_result[canonical_interface_name_edited(parent_lag)] = (
                        canonical_interface_name_list_edited(
                            properties.get("port_channel").get(
                                "port_channel_member_intfs"
                            )
                        )
                    )
    else:
        raise NotImplementedError
        # TODO Other platforms

    return converted_result


def convert_interfaces_ips(napalm_interfaces_ip, platform="iosxr"):
    """
    Returns all configured IP addresses on all interfaces as a dictionary of dictionaries.
    Keys of the main dictionary represent the name of the interface.
    Values of the main dictionary represent are dictionaries that may consist of two keys
    'ipv4' and 'ipv6' (one, both or none) which are themselves dictionaries with the IP
    addresses as keys.
    Each IP Address dictionary has the following keys:
        * prefix_length (int)
    Example::
        {
            u'FastEthernet8': {
                u'ipv4': {
                    u'10.66.43.169': {
                        'prefix_length': 22
                    }
                }
            },
            u'Loopback555': {
                u'ipv4': {
                    u'192.168.1.1': {
                        'prefix_length': 24
                    }
                },
                u'ipv6': {
                    u'1::1': {
                        'prefix_length': 64
                    },
                    u'2001:DB8:1::1': {
                        'prefix_length': 64
                    },
                    u'2::': {
                        'prefix_length': 64
                    },
                    u'FE80::3': {
                        'prefix_length': u'N/A'
                    }
                }
            },
            u'Tunnel0': {
                u'ipv4': {
                    u'10.63.100.9': {
                        'prefix_length': 24
                    }
                }
            }
        }

    # Extra! Convert all addresses to lower case
    """
    parsed_result = {}

    for ifc, params in napalm_interfaces_ip.items():
        ipv4 = params.get("ipv4", None)
        ipv6 = params.get("ipv6", None)

        # WARNING! - convert all adresses (ipv6) to lower case!!!
        if ipv4:
            lower_ipv4 = dict((k.lower(), v) for k, v in ipv4.items())
        if ipv6:
            lower_ipv6 = dict((k.lower(), v) for k, v in ipv6.items())

        canonical_ifc = canonical_interface_name_edited(ifc)

        if ipv4 or ipv6:
            parsed_result[canonical_ifc] = {}
            ipv4 and parsed_result[canonical_ifc].update({"ipv4": lower_ipv4})
            ipv6 and parsed_result[canonical_ifc].update({"ipv6": lower_ipv6})

    return parsed_result


def convert_interfaces_vlans_iosxr(genie_show_interfaces):
    # Converts subinterfaces to vlans
    # DEPRECATED - will be removed
    """return FORMAT
    {'vlans': {'1': {'interfaces': ['GigabitEthernet0/2/5',
                                    'GigabitEthernet0/2/23'],
                    'shutdown': False,
                    'state': 'active',
                    'vlan_id': '1'},
            '10': {'interfaces': ['GigabitEthernet0/2/0',
                                    'GigabitEthernet0/2/4'],
                    'shutdown': False,
                    'state': 'active',
                    'vlan_id': '10'},
            '2': {'interfaces': [],
                    'shutdown': False,
                    'state': 'unsupport',
                    'vlan_id': '2'}},
    'vlans_interfaces': {'GigabitEthernet0/2/0': {'vlan': '10'},
                        'GigabitEthernet0/2/9': {'vlan': '1'}},
    'vlans_trunks': {}
    }
    """

    RESULT = {"vlans": None, "vlans_interfaces": None, "vlans_trunks": None}
    xr_vlans = {}
    xr_vlans_interfaces = {}
    vlans_trunks = {}

    for interface, properties in genie_show_interfaces.items():
        interface = canonical_interface_name_edited(interface)
        encap1 = properties.get("encapsulations")
        if encap1:
            encap2 = encap1.get("encapsulation")
            if encap2 == "dot1q":
                vid = interface.split(".")[1]
                xr_vlans[vid] = {
                    "interfaces": [interface],
                    "mtu": properties.get("mtu"),
                    "name": f"vlan-{vid}",
                    "shutdown": not properties.get("enabled"),
                    "state": "active",
                    "vlan_id": vid,
                }

                xr_vlans_interfaces[interface] = {"vlan": vid}

    RESULT.update(
        {
            "vlans": xr_vlans,
            "vlans_interfaces": xr_vlans_interfaces,
            "vlans_trunks": vlans_trunks,
        }
    )

    return RESULT


def convert_interfaces_vlans_nxos(show_interface_switchport, show_vlan):
    RESULT = {"vlans": None, "vlans_interfaces": None, "vlans_trunks": None}

    converted_vlans = {}
    if show_vlan.get("vlans"):
        for vid, vlan_params in show_vlan.get("vlans").items():
            converted_vlans[vid] = {
                "vlan_id": vlan_params.get("vlan_id"),
                "name": vlan_params.get("name"),
                "shutdown": vlan_params.get("shutdown"),
                "state": vlan_params.get("state"),
                "interfaces": canonical_interface_name_list_edited(
                    vlan_params.get("interfaces", "")
                ),
            }

    RESULT["vlans"] = converted_vlans

    interfaces = {}
    for ifc_name, values in show_interface_switchport.items():
        can_ifc_name = canonical_interface_name_edited(ifc_name)
        if values["switchport_mode"] == "access" and str(values["access_vlan"]) != 1:
            interfaces[can_ifc_name] = {}
            interfaces[can_ifc_name]["status"] = values["switchport_status"]
            interfaces[can_ifc_name]["name"] = values["access_vlan_mode"]
            interfaces[can_ifc_name]["vlan"] = values["access_vlan"]
        elif values["switchport_mode"] == "trunk":
            interfaces[can_ifc_name] = {}
            interfaces[can_ifc_name]["status"] = values["switchport_status"]
            interfaces[can_ifc_name]["name"] = values["access_vlan_mode"]
            interfaces[can_ifc_name]["vlan"] = "trunk"

    RESULT["vlans_interfaces"] = interfaces

    trunks = {}
    for ifc_name, values in show_interface_switchport.items():
        can_ifc_name = canonical_interface_name_edited(ifc_name)
        if values["switchport_mode"] == "trunk" and values.get("trunk_vlans"):
            trunks[can_ifc_name] = {}
            trunks[can_ifc_name]["encapsulation"] = values[
                "admin_priv_vlan_trunk_encapsulation"
            ]
            trunks[can_ifc_name]["native_vlan"] = values["native_vlan"]
            trunks[can_ifc_name]["status"] = values["switchport_status"]
            trunks[can_ifc_name]["name"] = can_ifc_name
            trunks[can_ifc_name]["vlan_list"] = vlanconfig_to_list_edited(
                values.get("trunk_vlans")
            )

    RESULT["vlans_trunks"] = trunks
    return RESULT


def convert_interfaces_vlans_iosxe(vlans, vlans_interfaces, trunks):
    """return FORMAT
    {'vlans': {'1': {'interfaces': ['GigabitEthernet0/2/5',
                                    'GigabitEthernet0/2/23'],
                    'shutdown': False,
                    'state': 'active',
                    'vlan_id': '1'},
            '10': {'interfaces': ['GigabitEthernet0/2/0',
                                    'GigabitEthernet0/2/4'],
                    'shutdown': False,
                    'state': 'active',
                    'vlan_id': '10'},
            '2': {'interfaces': [],
                    'shutdown': False,
                    'state': 'unsupport',
                    'vlan_id': '2'}},
    'vlans_interfaces': {'GigabitEthernet0/2/0': {'status': 'connected','vlan': '10'},
                        'GigabitEthernet0/2/16': {'status': 'connected','vlan': 'trunk'},
                        'GigabitEthernet0/2/17': {'status': 'notconnect','vlan': '1'},
                        'GigabitEthernet0/2/9': {'status': 'notconnect','vlan': '1'}},
    'vlans_trunks': {'GigabitEthernet0/2/16': {'encapsulation': '802.1q',
                                                'mode': 'on',
                                                'name': 'GigabitEthernet0/2/16',
                                                'status': 'trunking',
                                                'vlan_list': [1, 2, 10, 11, 12]}}}
    """

    converted_vlans = {}
    if vlans.get("vlans"):
        for vid, vlan_params in vlans.get("vlans").items():
            converted_vlans[vid] = {
                "vlan_id": vlan_params.get("vlan_id"),
                "name": vlan_params.get("name"),
                "shutdown": vlan_params.get("shutdown"),
                "state": vlan_params.get("state"),
                "interfaces": canonical_interface_name_list_edited(
                    vlan_params.get("interfaces", "")
                ),
            }

    converted_vlans_interfaces = {}
    if vlans_interfaces.get("interfaces"):
        for ifc, params in vlans_interfaces.get("interfaces").items():
            ifc_name = canonical_interface_name_edited(ifc)
            converted_vlans_interfaces[ifc_name] = {
                "status": params.get("status"),
                "vlan": params.get("vlan"),
            }

    converted_trunks = {}
    if trunks.get("interface"):
        for ifc, params in trunks.get("interface").items():
            ifc_name = canonical_interface_name_edited(ifc)
            converted_trunks[ifc_name] = {
                "name": params.get("name"),
                "mode": params.get("mode"),
                "encapsulation": params.get("encapsulation"),
                "status": params.get("status"),
                "vlan_list": vlanconfig_to_list_edited(
                    params.get("vlans_allowed_active_in_mgmt_domain")
                ),
            }

    return {
        "vlans": converted_vlans,
        "vlans_interfaces": converted_vlans_interfaces,
        "vlans_trunks": converted_trunks,
    }


def convert_bandwidth_interfaces(genie_interfaces, platform="iosxr"):
    converted_result = {}
    """Example of Netbox friendly format
    
    {
        'canonical_interface_name':  10000 #bandwidth
        'canonical_interface_name' 110000000000 #bandwidth
    }
    """

    for ifc_name, properties in genie_interfaces.items():
        canonical_ifc_name = canonical_interface_name_edited(ifc_name)
        if platform in ["nxos", "nxos_ssh"]:
            pass
        elif platform in ["iosxr"]:
            bandwidth = properties.get("bandwidth", None)
            converted_result[canonical_ifc_name] = bandwidth
        elif platform in ["iosxe", "ios"]:
            bandwidth = properties.get("bandwidth", None)
            converted_result[canonical_ifc_name] = bandwidth
        else:
            raise NotImplementedError
        # TODO Other platforms

    return converted_result



from genie.testbed import load
from pyats.contrib.creators.netbox import Netbox

from netbox_network_importer.config import get_config


class PyAtsNetbox:
    def __init__(
        self,
        testbed_dict=None,
        topology=False,
        url_filter="platform_id__n=null&has_primary_ip=True",
    ):
        """Create PyATS testbed structure through Netbox

        The value of the temp parameter is stored as a value in
        the class variable temperature. The given value is converted
        into a float value if not yet done.

        :param testbed_dict: pass custom testbed configuration, default: Connection to netbox
        :param topology: Loads tons of information from Netbox, takes a lot of time. default: False
        """

        if not testbed_dict:
            testbed_dict = Netbox(
                netbox_url=get_config()["netbox"]["NETBOX_INSTANCE_URL"],
                user_token=get_config()["netbox"]["NETBOX_API_TOKEN"],
                def_user=get_config()["tacacs"]["TACACS_USERNAME"],
                def_pass=get_config()["tacacs"]["TACACS_PASSWORD"],
                encode_password=True,
                topology=topology,  # Dont load IP addresses etc
                url_filter=url_filter,
                custom_data_source=[],
            )

        self.testbed = load(testbed_dict._generate())

    def get_device(self, hostname):
        return self.testbed.devices[hostname]

    def get_testbed(self):
        return self.testbed

    def connect_device(
        self,
        hostname,
        init_exec_commands=[
            "terminal length 0 ",
        ],
        init_config_commands=[],
        log_stdout=False,
        learn_hostname=True,
        learn_os=True,
    ):
        device = self.get_device(hostname)
        device.connect(
            init_exec_commands=init_exec_commands,
            init_config_commands=init_config_commands,
            log_stdout=log_stdout,
            learn_hostname=learn_hostname,
            learn_os=learn_os,
        )

        return device

    def get_devices(self):
        return self.testbed.devices

    def connection_test(self, hostname):
        try:
            self.connect_device(hostname)
            return (True, f"{hostname} connected successfuly")
        except Exception as e:
            return (False, e)



"""Package configuration."""
from setuptools import find_packages, setup
from netbox_network_importer import __appname__

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

requirements = [
    'click',
    'pynetbox==7.4.1',
    'python-slugify',
    'appdirs',
    'pyyaml',
    'python-dotenv',
    'pyats[full]==25.2',
    'netutils',
    'nornir==3.5.0',
    'nornir-netbox',
    'nornir_utils',
    'nornir_rich',
    'deepdiff',
    'napalm==5.0.0',
    'ncclient',
    'loguru',
    'dictdiffer',
    'json2html',
    'jsonpickle',
]

setup_requirements = [
    "pytest-runner",
]

test_requirements = [
    "pytest", "pytest-cov", "pytest-mock", "pynxos", "ipdb"
]

setup(
    name=__appname__,
    author="Jan Krupa",
    author_email="jan.krupa@cesnet.cz",
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "Intended Audience :: Developers",
        "Development Status :: 4 - Beta",
        "Operating System :: Unix",
    ],
    description='Poll data from devices and store them into Netbox',
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    include_package_data=True,
    keywords="netbox,network",
    packages=find_packages(),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://gitlab.cesnet.cz/701/done/netbox_network_importer",
    version='1.1.6',
    zip_safe=False,
    python_requires='>=3.6, <4',
    entry_points={
        'console_scripts': [
            'netbox_network_importer=netbox_network_importer.__main__:cli',
        ]
    },
)



from netbox_network_importer.connections.Netbox import Netbox
from netbox_network_importer.helper import get_diff
from netbox_network_importer.results.results import NetboxResult, Status, Action


def process_interfaces_vlans(host, parsed_data) -> list([NetboxResult]):
    RESULTS = []  # list of NetboxResult

    # Init connection to Netbox via PyNetbox
    netbox = Netbox.connect()

    # find the device instance in netbox
    netbox_device = netbox.dcim.devices.get(host.data['id'])

    # INIT Dictionary of stored vlans in NETBOX
    # if configured vlan is found, it will be removed from dict
    # vlans left in dictionary will be removed at the end
    netbox_interfaces_with_vlan = get_interfaces_with_vlan(
        netbox_conn=netbox, device_id=netbox_device.id)

    vlans_interfaces = parsed_data['vlans_interfaces']
    vlans_trunks = parsed_data['vlans_trunks']
    vlans = parsed_data['vlans']

    for interface_name, ifc_vlan_data in vlans_interfaces.items():
        netbox_ifc = netbox.dcim.interfaces.get(
            device_id=netbox_device.id, name__ie=interface_name)

        # skip if interface is not found in netbox
        if not netbox_ifc:
            RESULTS.append(
                NetboxResult(
                    result=f"Interface {interface_name} NOT FOUND in Netbox - check interface configuration on device",
                    status=Status.CHECK_MANUALLY,
                    action=Action.LOOKUP,
                    diff=ifc_vlan_data,
                )
            )
            continue

        # Skip IGNORED interfaces
        if netbox_ifc.custom_fields.get("ignore_importer", False) == True:
            RESULTS.append(NetboxResult(
                result=f"{netbox_ifc.name} - Ignored by Importer - Skipping", status=Status.SKIPPED, action=Action.LOOKUP, diff=""))
            continue

        # Assign multiple tagged vlan to interface
        if ifc_vlan_data['vlan'] == 'trunk':
            netbox_vlans = []

            # if interface is not found in configured TRUNKS
            if not vlans_trunks.get(interface_name):
                if netbox_ifc.lag != None:
                    NetboxResult(
                        result=f"Device: {host.name} Interface: {interface_name} mode: trunk. Part of port channel",
                        status=Status.NOT_CHANGED,
                        action=Action.LOOKUP,
                        diff=None
                    )
                    continue
                elif ifc_vlan_data.get("status", None) == 'disabled' or ifc_vlan_data.get("status", None) == 'notconnect':
                    continue
                else:
                    #RESULTS.append(
                    #    NetboxResult(
                    #        result=f"Device: {host.name} Interface: {interface_name} mode: trunk. Unable to parse trunks from device",
                    #        status=Status.CHECK_MANUALLY,
                    #        action=Action.LOOKUP,
                    #        diff={"vlan_data": ifc_vlan_data,
                    #              "trunks": vlans_trunks}
                    #    )
                    #)
                    continue

            vlans_vids_allowed_on_trunk = vlans_trunks[interface_name]['vlan_list']

            # Go through each vlan vid in trunk mode and assing it
            for vid_integer in vlans_vids_allowed_on_trunk:
                # convert vlan_vid int to string
                vid = str(vid_integer)

                # TODO: what if name does not exists
                vname = vlans.get(vid, {}).get('name')
                if not vname:
                    # TODO: Ignore vid;s which does not have crated VLANs
                    #RESULTS.append(
                    #    NetboxResult(
                    #        result=f"Device: {host.name} Interface: {interface_name} has configured VLAN ID: {vid}. But VLAN does not exists on Device.",
                    #        status=Status.CHECK_MANUALLY,
                    #        action=Action.LOOKUP,
                    #        diff={"host": host.name, "ifc": interface_name,
                    #              "vid": vid, "vlan_vid_data": vlans.get(vid, None)}
                    #    )
                    #)
                    continue

                # Skip VID == 1, skip 999?
                if vid == '1':
                    continue

                try:
                    RESULTS.append(create_vlan(
                        vname=vname, vid=vid, netbox_conn=netbox))
                    netbox_vlan = netbox.ipam.vlans.get(
                        vid=vid, name=vname, site_id='null')

                    netbox_vlans.append(netbox_vlan)
                except Exception as e:
                    RESULTS.append(
                        NetboxResult(
                            result=f"Unable to get/create vlan {vname} {vid} - Processing VLANs is aborted",
                            status=Status.EXCEPTION,
                            action=Action.LOOKUP,
                            exception=e,
                            diff={}
                        )
                    )
                    # end processing
                    return RESULTS

                # remove vlan from already configured vlans on interface (if exists)
                # vlans left in alredy configured vlans will be removed
                netbox_interfaces_with_vlan = remove_configured_vlan(
                    configured_vlan=netbox_interfaces_with_vlan,
                    ifc_name=netbox_ifc.name,
                    vid=netbox_vlan.vid,
                    vname=vname,
                    vlan_type='tagged_vlans')

            # assign VLAN to interface
            RESULTS.append(add_tagged_vlans_to_interface(
                netbox_ifc=netbox_ifc, netbox_vlans=netbox_vlans))

        elif ifc_vlan_data['vlan'] == 'routed':
            if ifc_vlan_data.get("status", None) == 'disabled' or ifc_vlan_data.get("status", None) == 'notconnect':
                continue
            else:
                RESULTS.append(
                    NetboxResult(
                        result=f"IFC: {netbox_ifc.name}, type {ifc_vlan_data['vlan']} - SKIPPING",
                        status=Status.NOT_CHANGED,
                        action=Action.LOOKUP,
                        diff={}
                    )
                )
                continue

        elif ifc_vlan_data['vlan'] in ['unassigned', 'unassigne']:
            continue

        # Assign untagged vlan to interface
        else:
            vid = str(ifc_vlan_data['vlan'])
            vname = vlans.get(vid, {}).get('name')
            if not vname:
                RESULTS.append(
                    NetboxResult(
                        result=f"Device: {host.name} Interface: {interface_name} has configured VLAN ID: {vid}. But VLAN does not exists on Device.",
                        status=Status.CHECK_MANUALLY,
                        action=Action.LOOKUP,
                        diff={"host": host.name, "ifc": interface_name,
                              "vid": vid, "vlan_vid_data": vlans.get(vid, None)}
                    )
                )
                continue

            if vid == '1':
                continue

            try:
                RESULTS.append(create_vlan(
                    vname=vname, vid=vid, netbox_conn=netbox))
                netbox_vlan = netbox.ipam.vlans.get(
                    vid=vid, name=vname, site_id='null')
            except Exception as e:
                RESULTS.append(
                    NetboxResult(
                        result=f"Unable to get/create vlan {vname} {vid} - Processing VLANs is aborted",
                        status=Status.EXCEPTION,
                        action=Action.LOOKUP,
                        exception=e,
                        diff={}
                    )
                )
                # end processing
                return RESULTS

            # add vlan to interface
            RESULTS.append(add_untagged_vlan_to_interface(
                netbox_ifc=netbox_ifc, netbox_vlan=netbox_vlan))

            # remove vlan from dictionary of already exisitng vlans in netbox (if exists)
            netbox_interfaces_with_vlan = remove_configured_vlan(configured_vlan=netbox_interfaces_with_vlan,
                                                                 ifc_name=netbox_ifc.name,
                                                                 vid=netbox_vlan.vid,
                                                                 vname=vname,
                                                                 vlan_type='untagged_vlan')

    # REMOVE vlan from interfaces which are not configured anymore
    # all vlans left in netbox_interfaces_with_vlan
    RESULTS.extend(delete_netbox_obsolete_vlans(
        vlans_to_remove=netbox_interfaces_with_vlan, netbox_device=netbox_device, netbox=netbox))
    return RESULTS


def add_tagged_vlans_to_interface(netbox_ifc, netbox_vlans):
    before = get_changes_tagged_vlans(netbox_ifc)

    for netbox_vlan in netbox_vlans:
        if not netbox_vlan in netbox_ifc.tagged_vlans:
            netbox_ifc.mode = 'tagged'
            netbox_ifc.tagged_vlans.append(netbox_vlan)

    after = get_changes_tagged_vlans(netbox_ifc)
    if netbox_ifc.save():
        return NetboxResult(status=Status.CHANGED,
                            diff=get_diff(before, after),
                            action=Action.CREATE,
                            result=f"{netbox_ifc.name} mode {netbox_ifc.mode} and VLANs ({netbox_vlans}) saved")
    else:
        return NetboxResult(status=Status.NOT_CHANGED,
                            diff=get_diff(before, after),
                            action=Action.CREATE,
                            result=f"{netbox_ifc.name} mode {netbox_ifc.mode}. VLANs already exists ({netbox_ifc.tagged_vlans})")


def remove_tagged_vlans_from_interface(netbox_vlans, netbox_ifc):
    before = get_changes_tagged_vlans(netbox_ifc)

    for identifier, nb_vlan in netbox_vlans.items():
        if nb_vlan in netbox_ifc.tagged_vlans:
            netbox_ifc.tagged_vlans.remove(nb_vlan)

    if not netbox_ifc.untagged_vlan and not netbox_ifc.tagged_vlans:
        netbox_ifc.mode = ""

    if netbox_ifc.save():
        netbox_ifc = netbox_ifc.api.dcim.interfaces.get(netbox_ifc.id)
        return NetboxResult(
            action=Action.DELETE,
            status=Status.CHANGED,
            result=f"Netbox tagged vlans {list(netbox_vlans.keys())} were removed from interface {netbox_ifc.name}",
            diff={"remove": [{"vid": vlan[0], "name": vlan[1]}
                             for vlan in netbox_vlans]}
        )
    else:
        return NetboxResult(
            action=Action.DELETE,
            status=Status.CHECK_MANUALLY,
            result=f"Unable to remove tagged vlans {netbox_vlans.keys()} from interface {netbox_ifc.name}",
            diff={}
        )


def add_untagged_vlan_to_interface(netbox_ifc, netbox_vlan):
    if netbox_vlan:
        before = get_changes_untagged_vlan(netbox_ifc)
        netbox_ifc.mode = 'access'
        netbox_ifc.untagged_vlan = netbox_vlan
    else:
        return NetboxResult(
            status=Status.FAILED,
            action=Action.LOOKUP,
            diff={},
            result=f"Vlan not found in netbox {netbox_ifc}, {netbox_vlan}")

    after = get_changes_untagged_vlan(netbox_ifc)
    if netbox_ifc.save():
        return NetboxResult(status=Status.CHANGED,
                            diff=get_diff(before, after),
                            action=Action.CREATE,
                            result=f"{netbox_ifc.name} mode {netbox_ifc.mode} and VLAN ({netbox_vlan.display}) saved")
    else:
        return NetboxResult(status=Status.NOT_CHANGED,
                            diff=get_diff(before, after),
                            action=Action.CREATE,
                            result=f"{netbox_ifc.name} mode {netbox_ifc.mode} and VLAN ({netbox_vlan.display}) already exists")


def remove_untagged_vlan_from_interface(netbox_vlan, netbox_ifc):
    before = get_changes_untagged_vlan(netbox_ifc)

    if netbox_ifc.untagged_vlan == netbox_vlan:
        netbox_ifc.untagged_vlan = None
        if not netbox_ifc.untagged_vlan and not netbox_ifc.tagged_vlans:
            netbox_ifc.mode = None

        if netbox_ifc.save():
            reload_ifc = netbox_ifc.api.dcim.interfaces.get(netbox_ifc.id)
            after = get_changes_untagged_vlan(reload_ifc)

            return NetboxResult(
                result=f"Untagged vlan: {netbox_vlan.display} removed from {netbox_ifc.name} - {netbox_ifc.device.name}",
                status=Status.CHANGED,
                action=Action.DELETE,
                exception=None,
                diff=get_diff(before, after)
            )
        else:
            return NetboxResult(
                result=f"Untagged vlan: {netbox_vlan.display} could not be removed from {netbox_ifc.name} - {netbox_ifc.device.name}",
                status=Status.ANOMALLY,
                action=Action.DELETE,
                exception=None,
                diff={}
            )
    else:
        return NetboxResult(
            result=f"Untagged vlan: {netbox_vlan.display} was already changed to {netbox_ifc.untagged_vlan.display} on {netbox_ifc.name} - {netbox_ifc.device.name}",
            status=Status.NOT_CHANGED,
            action=Action.UPDATE,
            exception=None,
            diff={}
        )


def create_vlan(vname, vid, netbox_conn):
    try:
        # lookup for vlan in netbox,
        # could raises error when multiple vlans are returned
        nb_vlan = netbox_conn.ipam.vlans.get(
            vid=vid, name=vname, site_id='null')
    except Exception as e:
        raise e

    if not nb_vlan:
        try:
            if netbox_conn.ipam.vlans.create({'vid': vid, 'name': vname}):
                return NetboxResult(status=Status.CHANGED,
                                    diff={'vid': vid, 'name': vname},
                                    action=Action.CREATE,
                                    result=f"Creating vlan ID: {vid} Name: {vname}")
            else:
                return NetboxResult(status=Status.ANOMALLY,
                                    diff={},
                                    action=Action.CREATE,
                                    result=f"Unable to create - Vlan ID: {vid} Name: {vname}")
        except Exception as e:
            # if request error on create, try to find again !!!
            # WORKAROUND: sometimes other thread creates a vlan in netbox before this action
            nb_vlan = netbox_conn.ipam.vlans.get(
                vid=vid, name=vname, site_id='null')
            if nb_vlan:
                return NetboxResult(status=Status.NOT_CHANGED,
                                    diff={},
                                    action=Action.CREATE,
                                    result=f"Vlan ID: {vid} Name: {vname} already exists")
            else:
                # TODO:
                raise e
    else:
        return NetboxResult(status=Status.NOT_CHANGED,
                            diff={},
                            action=Action.CREATE,
                            result=f"Vlan ID: {vid} Name: {vname} already exists")


def delete_netbox_obsolete_vlans(vlans_to_remove, netbox_device, netbox):
    RESULTS = []
    for ifc, type_vlans in vlans_to_remove.items():
        nb_ifc = netbox.dcim.interfaces.get(
            device_id=netbox_device.id, name=ifc)

        # Skip IGNORED interfaces
        if nb_ifc.custom_fields.get("ignore_importer", False) == True:
            RESULTS.append(NetboxResult(
                result=f"{nb_ifc.name} - Ignored by Importer - Skipping", status=Status.SKIPPED, action=Action.DELETE, diff=""))
            continue

        if type_vlans.get('untagged_vlan'):
            nb_vlan = type_vlans.get('untagged_vlan')
            RESULTS.append(
                remove_untagged_vlan_from_interface(nb_vlan, nb_ifc))

        if type_vlans.get('tagged_vlans'):
            tagged_vlans = type_vlans.get('tagged_vlans')
            # Remove all tagged_vlans at once
            RESULTS.append(remove_tagged_vlans_from_interface(
                tagged_vlans, nb_ifc))

    return RESULTS


def remove_configured_vlan(configured_vlan, ifc_name, vid, vname, vlan_type):
    if vlan_type == 'untagged_vlan':
        if configured_vlan.get(ifc_name):
            if configured_vlan[ifc_name].get(vlan_type):
                if configured_vlan[ifc_name][vlan_type].vid == vid:
                    configured_vlan[ifc_name].pop(vlan_type)

    if vlan_type == 'tagged_vlans':
        if configured_vlan.get(ifc_name):
            if configured_vlan[ifc_name].get(vlan_type):
                configured_vlan[ifc_name][vlan_type].pop((vid, vname), None)

    return configured_vlan


def get_interfaces_with_vlan(netbox_conn, device_id):
    tagged = get_tagged_vlans_interfaces(netbox_conn, device_id)
    untagged = get_untagged_vlan_interfaces(netbox_conn, device_id)

    merged_dict = {}

    for ifc, vlans in tagged.items():
        if merged_dict.get(ifc, None):
            merged_dict[ifc].update(vlans)
        else:
            merged_dict[ifc] = vlans

    for ifc, vlan in untagged.items():
        if merged_dict.get(ifc, None):
            merged_dict[ifc].update(vlan)
        else:
            merged_dict[ifc] = vlan

    return merged_dict


def get_tagged_vlans_interfaces(netbox_conn, device_id):
    """ return FORMAT
        {'GigabitEthernet0/2/16': {'tagged_vlans': {(1, 'default'): PYNETBOX_OBJECT,
                                            (12, 'dwdm-T6'): PYNETBOX_OBJECT}}
    """
    res = {
        ifc.name: {"tagged_vlans": {(vlan.vid, vlan.name): vlan for vlan in [
            vlan for vlan in ifc.tagged_vlans]}}
        for ifc in netbox_conn.dcim.interfaces.filter(device_id=device_id) if ifc.tagged_vlans
    }

    return res


def get_untagged_vlan_interfaces(netbox_conn, device_id):
    res = {
        ifc.name: {"untagged_vlan": ifc.untagged_vlan}
        for ifc in netbox_conn.dcim.interfaces.filter(device_id=device_id) if ifc.untagged_vlan
    }

    return res


def get_changes_tagged_vlans(netbox_ifc):
    # return {vlan.vid: vlan.name for vlan in netbox_ifc.tagged_vlans}
    return [{"name": vlan.name, "vid": vlan.vid} for vlan in netbox_ifc.tagged_vlans]


def get_changes_untagged_vlan(netbox_ifc):
    if netbox_ifc.untagged_vlan:
        return {"vid": netbox_ifc.untagged_vlan.vid, "name": netbox_ifc.untagged_vlan.name}
    else:
        return {}



from netbox_network_importer.results.results import NetboxResult, Status, Action
from netbox_network_importer.helper import canonical_interface_name_edited, get_diff
from netbox_network_importer.connections.Netbox import Netbox


def process_interfaces_ips(host, parsed_interfaces) -> list:
    """CRUD on netbox interfaces IPs
    - Delete IPs, that exists in netbox but are not passed from device
    - Update IPs, which already exists in netbox
    - Create IPs, which does not exists in netbox

    :param parsed_interfaces: parsed interfaces for netbox operations
    """

    # Init connection to Netbox via PyNetbox
    nb = Netbox.connect()

    RESULTS = []

    # find the device instance in netbox
    dev = nb.dcim.devices.get(host.data['id'])

    # find interfaces linked to the device
    all_device_ifcs_recordset = nb.dcim.interfaces.filter(device_id=dev.id)

    # convert filtered interfaces into dictionary of pynetbox interface instances
    nb_interfaces_dict = {canonical_interface_name_edited(
        nb_ifc.name): nb_ifc for nb_ifc in all_device_ifcs_recordset}

    # Create DICT of ALL
    #    = parsed interfaces from genie (that with IP address assigned)
    #    = all interfaces stored in netbox (if not parsed with genie, empty dict of ip addresses passed)
    ALL_INTERFACES = {}
    ALL_INTERFACES.update({k: {} for k in nb_interfaces_dict})
    ALL_INTERFACES.update(parsed_interfaces)

    # Loop over all interfaces and process their IP addreses (CRUD)
    for ifc, parsed_ips in ALL_INTERFACES.items():
        # get NB interface instance
        nb_ifc = nb_interfaces_dict.get(ifc, None)
        # get NB interface ip addresses and convert to dictionary
        if nb_ifc == None:
            RESULTS.append(
                NetboxResult(
                    result=f"Interface {ifc} NOT FOUND!, Configured IPS: {parsed_ips} - Check manually",
                    status=Status.ANOMALLY,
                    action=Action.LOOKUP,
                    diff={}
                )
            )
            continue

        # Skip IGNORED interfaces
        if nb_ifc.custom_fields.get("ignore_importer", False) == True:
            RESULTS.append(NetboxResult(
                result=f"{nb_ifc.name} - Ignored by Importer - Skipping", status=Status.SKIPPED, action=Action.LOOKUP, diff=""))
            continue

        nb_interface_ips_recordset = nb.ipam.ip_addresses.filter(
            interface_id=nb_ifc.id)

        # key = ip/mask
        nb_interface_ips_dict = {
            ip.address: ip for ip in nb_interface_ips_recordset}

        # Go through different family addresses
        for address_family in parsed_ips:
            if address_family == 'ipv4':
                family = {"value": 4, "label": "IPv4"}
            elif address_family == 'ipv6':
                family = {"value": 6, "label": "IPv6"}

            for ip, props in parsed_ips[address_family].items():
                # TODO:
                if ip == '0.0.0.0':
                    continue

                prefix = props.get('prefix_length', 32)
                ip = f"{ip}/{prefix}"

                # Pop device from dictionary - if none, create it
                if not nb_interface_ips_dict.pop(ip, None):
                    ip_params = {
                        'assigned_object_type':  "dcim.interface",
                        'assigned_object_id': nb_ifc.id,
                        'family': family,
                        'address': ip
                    }
                    RESULTS.append(ip_create(ip_params=ip_params,
                                   nb=nb, ifc_name=nb_ifc.name))

        # TODO: DELETE!!
        # If any interface is left in `nb_interfaces_dict`, then it should be removed from netbox
        #   - Interface exists in netbox, but the interface was not passed from network
        for _k, nb_ip_address in nb_interface_ips_dict.items():
            # Skip IGNORED interfaces
            if nb_ifc.custom_fields.get("ignore_importer", False) == True:
                RESULTS.append(NetboxResult(
                    result=f"{ifc.name} - {nb_ip_address.address} - Ignored by Importer - Skipping", status=Status.SKIPPED, action=Action.DELETE, diff=""))
                continue
            RESULTS.append(ip_delete(ip=nb_ip_address, ifc_name=nb_ifc.name))

    return RESULTS


def ip_delete(ip, ifc_name):
    if ip.delete():
        return NetboxResult(
            result=f"{ifc_name} - {ip.address} - deleted successfully",
            status=Status.CHANGED,
            action=Action.DELETE,
            diff={}
        )
    else:
        return NetboxResult(result=f"{ifc_name} - {ip.address} - could not be deleted",
                            status=Status.FAILED,
                            action=Action.DELETE,
                            diff={}
                            )


def ip_create(ip_params, nb, ifc_name):
    netbox_ip = nb.ipam.ip_addresses.create(ip_params)

    if netbox_ip:
        return NetboxResult(
            result=f"{ifc_name} - {netbox_ip.address} - created successfully",
            status=Status.CHANGED,
            action=Action.CREATE,
            diff=ip_params
        )
    else:
        return NetboxResult(
            result=f"{ifc_name} - {ip_params['address']} - Unable to create",
            status=Status.FAILED,
            action=Action.CREATE,
            diff={}
        )



import pynetbox

from netbox_network_importer.config import get_config


class Netbox:
    def __init__(self):
        self.netbox_url = get_config()['netbox']["NETBOX_INSTANCE_URL"]
        self.api_token = get_config()['netbox']["NETBOX_API_TOKEN"]
        self.connection = None

    # https://pynetbox.readthedocs.io/en/latest/endpoint.html
    def open(self):
        try:
            # private_key_file='/path/to/private-key.pem',
            self.connection = pynetbox.api(
                self.netbox_url, token=self.api_token)
        except Exception as e:
            print("Could not establish connection to pynetbox")
            raise e

    def connect():
        return pynetbox.api(get_config()['netbox']["NETBOX_INSTANCE_URL"], get_config()['netbox']["NETBOX_API_TOKEN"])
